""" Forms class (via REST) """
from json import dump, load
import requests

from .pyeloqua import Eloqua
from .error_handling import _elq_error_

############################################################################
# Constant definitions
############################################################################

POST_HEADERS = {'Content-Type': 'application/json'}

############################################################################
# Form class
############################################################################


class Form(Eloqua):
    """ Extension for Form operations via REST API """

    def __init__(self, username=None, password=None, company=None,
                 test=False, form_id=None, form_path=None):
        """
        Initialize form class:

        Arguments:
        :param string username: Eloqua username
        :param string password: Eloqua password
        :param string company: Eloqua company instance
        :param bool test: Sets up test instance; does not connect to Eloqua
        :param int form_id: Eloqua Form ID; optional
        :param string form_path: Path to JSON file with Eloqua Form definition
        :return: Forms object
        """
        Eloqua.__init__(self, username, password, company, test)

        if form_id is None and form_path is None:
            raise ValueError('One of form_id or form_path is required')
        elif form_id is not None:
            self.form_id = form_id
            self.schema = self._get_form()
        elif form_path is not None:
            self.schema = self._load_form(form_path)
            self.form_id = self.schema['id']

        self.fields = self._get_fields()

    def _get_form(self):
        """ retrieve complete form definition """

        url = self.rest_base + 'assets/form/{id}?depth=complete'.format(
            id=self.form_id
        )

        req = requests.get(url=url, auth=self.auth)

        _elq_error_(req)

        return req.json()

    def _get_fields(self):
        """ get id-to-field mapping """
        fmap = {}
        # using this infernal if/then structure because more than fields are
        # included in 'elements'
        for field in self.schema['elements']:
            if isinstance(field, dict):
                if 'type' in field and field['type'] == 'FormField':
                    fmap[field['id']] = field['htmlName']

        return fmap

    def _load_form(self, form_path):
        """ load form definition from json file """

        return load(open(form_path, 'r'))

    def write_form(self, form_path):
        """ write form definition to json file """

        dump(self.schema, open(form_path, 'w'))

    def _parse_data(self, data):
        """ parse data set to pythonic list of dict """

        out_data = []

        for row in data:

            prow = {}

            for field in row['fieldValues']:

                if 'value' in field.keys():

                    prow[self.fields[field['id']]] = field['value']

            prow['id'] = row['id']
            prow['submittedAt'] = row['submittedAt']

            out_data.append(prow)

        return out_data


    def get_count(self, start, end):
        """
        retrieve count of form submissions

        :param datetime start: filter start unixtime
        :param datetime end: filter end unixtime
        """

        url = self.rest_base + \
            'data/form/{id}?startAt={start}&endAt={end}&count=1'.format(
                id=self.form_id,
                start=start,
                end=end
            )

        req = requests.get(url=url, auth=self.auth)

        _elq_error_(req)

        return req.json()['total']


    def get_data(self, start, end):
        """
        retrieve form submission data

        :param datetime start: filter start unixtime
        :param datetime end: filter end unixtime
        """

        url = self.rest_base + \
            'data/form/{id}?startAt={start}&endAt={end}&page={page}'

        has_more = True
        page = 1
        results = []

        while has_more:

            url_page = url.format(
                id=self.form_id,
                start=start,
                end=end,
                page=page
            )

            req = requests.get(url=url_page, auth=self.auth)

            _elq_error_(req)

            results.extend(self._parse_data(req.json()['elements']))

            page += 1
            has_more = len(results) < req.json()['total']

        return results



""" pyeloqua is a package of wrappers functions for interacting with Eloqua over Bulk API 2.0.
    Right now, they cover Contacts, Accounts, and Custom Data Objects."""

from .pyeloqua import Eloqua
from .bulk import Bulk
from .forms import Form
from .system_fields import ACTIVITY_FIELDS, CONTACT_SYSTEM_FIELDS, ACCOUNT_SYSTEM_FIELDS
from .error_handling import EloquaBulkSyncTimeout, EloquaValidationError



""" Eloqua.Form unit tests """

from json import dump, load
from nose.tools import raises
from mock import patch, Mock

from pyeloqua import Form

###############################################################################
# Constants
###############################################################################

POST_HEADERS = {'Content-Type': 'application/json'}

FORM_SAMPLE = {
    'type': 'Form',
    'currentStatus': 'Draft',
    'id': '18',
    'createdAt': '1435261611',
    'createdBy': '19',
    'depth': 'complete',
    'folderId': '7',
    'name': 'Test Form',
    'permissions': 'fullControl',
    'updatedAt': '1435261612',
    'updatedBy': '19',
    'elements': [
        {
            'type': 'FormField',
            'id': '174',
            'name': 'Email Address',
            'createdFromContactFieldId': '100001',
            'dataType': 'text',
            'displayType': 'text',
            'fieldMergeId': '1',
            'htmlName': 'emailAddress'
        },
        {
            'type': 'FormField',
            'id': '175',
            'name': 'First Name',
            'createdFromContactFieldId': '100002',
            'dataType': 'text',
            'displayType': 'text',
            'fieldMergeId': '1',
            'htmlName': 'firstName'
        },
        {
            'type': 'FormField',
            'id': '176',
            'name': 'Submit',
            'dataType': 'text',
            'displayType': 'submit',
            'htmlName': 'submit'
        }
    ]
}

FORM_FIELDS = {
    '174': 'emailAddress',
    '175': 'firstName',
    '176': 'submit'
}

TEST_DATA = {
    'elements': [
        {
            'id': '1',
            'submittedAt': '1234567899',
            'fieldValues': [
                {
                    'id': '174',
                    'value': 'test1@test.com'
                },
                {
                    'id': '175',
                    'value': 'john'
                }
            ]
        },
        {
            'id': '2',
            'submittedAt': '1234567899',
            'fieldValues': [
                {
                    'id': '174',
                    'value': 'test2@test.com'
                },
                {
                    'id': '175'
                }
            ]
        }
    ],
    'page': 1,
    'pageSize': 1000,
    'total': 2
}

PARSED_DATA = [
    {
        'id': '1',
        'submittedAt': '1234567899',
        'emailAddress': 'test1@test.com',
        'firstName': 'john'
    },
    {
        'id': '2',
        'submittedAt': '1234567899',
        'emailAddress': 'test2@test.com'
    }
]

###############################################################################
# Tests - initialization
###############################################################################


@patch('pyeloqua.forms.requests.get')
def test_f_instance(mock_get):
    """ Initialize Form via form_id """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = FORM_SAMPLE
    assert isinstance(Form(test=True, form_id=1), Form)


@patch('pyeloqua.forms.requests.get')
def test_f_init_id(mock_get):
    """ Init Form instance with form_id """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = FORM_SAMPLE
    assert Form(test=True, form_id=1) is not None


def test_f_init_path():
    """ Init Form instance with form_path """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    assert Form(test=True, form_path=fpath) is not None


def test_f_init_path_id():
    """ Init Form instance with form_path and sets id """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    assert Form(test=True, form_path=fpath).form_id == '18'


def test_f_init_schema():
    """ Init form, set schema """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    assert Form(test=True, form_path=fpath).schema == FORM_SAMPLE


def test_f_init_fields():
    """ Init form, set fields """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    assert Form(test=True, form_path=fpath).fields == FORM_FIELDS


@raises(ValueError)
def test_f_id_path_req():
    """ Init form requires form_id or form_path """
    Form(test=True)


###############################################################################
# Tests - helpers
###############################################################################

def test_f_writeform():
    """ Init Form instance with form_path """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    form = Form(test=True, form_path=fpath)
    form.write_form('/tmp/pyeloqua_test_f_writeform.json')
    ofile = load(open('/tmp/pyeloqua_test_f_writeform.json', 'r'))
    assert ofile == form.schema


@patch('pyeloqua.forms.requests.get')
def test_f_get_form_calls(mock_get):
    """ _get_form calls API correctly """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = FORM_SAMPLE
    form = Form(test=True, form_id=1)
    url = form.rest_base + 'assets/form/1?depth=complete'
    mock_get.assert_called_with(url=url, auth=form.auth)


def test_f_parse_data():
    """ _parse_data returns expected dict """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    form = Form(test=True, form_path=fpath)
    out = form._parse_data(TEST_DATA['elements'])
    assert out == PARSED_DATA


@patch('pyeloqua.forms.requests.get')
def test_f_get_data_parse(mock_get):
    """ get_data returns complete parsed data """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    form = Form(test=True, form_path=fpath)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = TEST_DATA
    out = form.get_data(start='123456789', end='987654321')
    assert out == PARSED_DATA


@patch('pyeloqua.forms.requests.get')
def test_f_get_data_calls(mock_get):
    """ get_data calls correct endpoint """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    form = Form(test=True, form_path=fpath)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = TEST_DATA
    form.get_data(start='123456789', end='987654321')
    url = form.rest_base + 'data/form/18?startAt=123456789&endAt=987654321&page=1'
    mock_get.assert_called_with(url=url, auth=form.auth)


@patch('pyeloqua.forms.requests.get')
def test_f_get_count_return(mock_get):
    """ get_count returns correct record count """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    form = Form(test=True, form_path=fpath)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = TEST_DATA
    out = form.get_count(start='123456789', end='987654321')
    assert out == 2


@patch('pyeloqua.forms.requests.get')
def test_f_get_count_calls(mock_get):
    """ get_count calls correct endpoint """
    fpath = '/tmp/pyeloqua_test_f_init_path.json'
    dump(FORM_SAMPLE, open(fpath, 'w'))
    form = Form(test=True, form_path=fpath)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = TEST_DATA
    form.get_count(start='123456789', end='987654321')
    url = form.rest_base + 'data/form/18?startAt=123456789&endAt=987654321&count=1'
    mock_get.assert_called_with(url=url, auth=form.auth)



from datetime import datetime
import json
import time
import warnings
import requests

from . import system_fields

API_VERSION = '2.0'

POST_HEADERS = {'Content-Type': 'application/json'}


class Eloqua(object):

    """ Eloqua instance:
            Wraps an Eloqua session to more easily access base API URLs and
            make instance and user info available.
            User must have at least 'Advanced Marketing User' or
            'API User' permissions
            Any contact level security labels will also apply through the API
    """

    def __init__(self, username=None, password=None, company=None,
                 test=False):
        """
        Create an Eloqua session given the following arguments:

        Arguments:

        :param string username: Eloqua username
        :param string password: Eloqua password
        :param string company: Eloqua company instance
        :param bool test: Sets up test instance; does not connect to Eloqua
        """

        if not test:

            if all(arg is not None for arg in (username, password, company)):

                url = 'https://login.eloqua.com/id'
                req = requests.get(url, auth=(company + '\\' + username,
                                              password))

                # check if any error codes
                req.raise_for_status()

                if req.json() == 'Not authenticated.':
                    raise ValueError('Invalid login credentials')
                else:
                    self.username = username
                    self.password = password
                    self.company = company
                    self.auth = (company + '\\' + username, password)
                    self.userId = req.json()['user']['id']
                    self.user_id = req.json()['user']['id']
                    self.userDisplay = req.json()['user']['displayName']
                    self.user_display = req.json()['user']['displayName']
                    self.urlBase = req.json()['urls']['base']
                    self.url_base = req.json()['urls']['base']
                    self.siteId = req.json()['site']['id']
                    self.site_id = req.json()['site']['id']

                    self.rest_bs_un = req.json()['urls']['apis'][
                        'rest']['standard']
                    self.restBase = self.rest_bs_un.format(
                        version=API_VERSION) #will deprecate this
                    self.rest_base = self.rest_bs_un.format(
                        version=API_VERSION)

                    self.bulk_bs_un = req.json()['urls']['apis']['rest']['bulk']
                    self.bulkBase = self.bulk_bs_un.format(
                        version=API_VERSION) #will deprecate this
                    self.bulk_base = self.bulk_bs_un.format(
                        version=API_VERSION)
            else:
                raise Exception(
                    'Please enter all required login details: company, username, password')

        else:
            self.username = 'test'
            self.password = 'test'
            self.company = 'test'
            self.auth = (self.company + '\\' + self.username, self.password)
            self.userId = 0
            self.userDisplay = 'Testing McTestFace'
            self.urlBase = "https://secure.p01.eloqua.com"
            self.siteId = 1
            self.restBase = "https://secure.p01.eloqua.com/API/REST/2.0/"
            self.bulkBase = "https://secure.p01.eloqua.com/API/BULK/2.0/"
            self.rest_base = "https://secure.p01.eloqua.com/API/REST/2.0/"
            self.bulk_base = "https://secure.p01.eloqua.com/API/BULK/2.0/"
            self.rest_bs_un = "https://secure.p01.eloqua.com/API/REST/{version}/"
            self.bulk_bs_un = "https://secure.p01.eloqua.com/API/BULK/{version}/"

    def set_bulk_version(self, version):
        """ update bulk version """

        self.bulk_base = self.bulk_bs_un.format(version=version)


    '''
        ###################################################
        Bulk API 2.0
        ###################################################
    '''

    def GetFields(self, entity, cdoID=0, fields=[]):
        """
            Retrieve all fields for a specified entitiy

            Arguments:

            * entity -- one of: contacts, customObjects, or accounts
            * cdoID -- identifier of specific CDO; required if
                entity = 'customObjects'; use method GetCdoId to retrieve
            * fields -- list of specific fields to retrieve,
                either by 'Display Name' or 'Database Name'; optional
        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if entity not in ['contacts', 'customObjects', 'accounts']:
            raise ValueError(
                "Please choose a valid 'entity' value: 'contacts', 'accounts', 'customObjects'")

        if entity == 'customObjects':
            if cdoID == 0:
                raise ValueError("Please specify a cdoID")

            url = self.bulkBase + '/customobjects/' + str(cdoID) + '/fields'
        else:
            url = self.bulkBase + entity + '/fields'

        req = requests.get(url, auth=self.auth)

        fieldsReturn = []

        if req.status_code == 200:

            if len(fields) > 0:

                for item in req.json()['items']:

                    if item['name'] in fields:
                        fieldsReturn.append(item)
                    elif item['internalName'] in fields:
                        fieldsReturn.append(item)

            else:

                fieldsReturn = req.json()['items']

            return fieldsReturn

        else:

            raise Exception("Failure getting fields: " + str(req.status_code))

    def CreateFieldStatement(self, entity, fields='', cdoID=0,
                             useInternalName=True, addSystemFields=[],
                             addActivityFields=[], activityType='',
                             leadScoreModelId=0, addSystemContactFields=[],
                             addLinkedContactFields=[], addLinkedAccountFields=[],
                             addAll=False):
        """
            Given a set of field names, create a "fields" statement for use in Bulk import/export definitions

            Arguments:

            * entity -- one of: contacts, customObjects, accounts, activities
            * fields -- list of specific fields to retrieve, either by 'Display Name' or 'Database Name'; optional
            * cdoID -- identifier of specific CDO; required if entity = 'customObjects'; use method GetCdoId to retrieve
            * useInternalName -- If True, import / export defined field names use 'Database Name'
            * addSystemFields -- DEPRECATED: Use addSystemContactFields instead. List of system fields to include in statement; see CONTACT_SYSTEM_FIELDS
            * addActivityFields -- List of activity fields to include; required if entity = 'activities'; see ACTIVITY_FIELDS
            * activityType -- export type
            * leadScoreModelId -- add lead score model fields to contact export
            * addSystemContactFields -- List of system fields to include in statement relative to [linked] contacts; see CONTACT_SYSTEM_FIELDS
            * addLinkedContactFields -- List of fields to add in CDO record exports
            * addLinkedAccountFields -- List of fields to add in Contact record exports
            * addAll -- ALL OF THE THINGS!!!!
        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if (entity in ['contacts', 'customObjects', 'accounts'] and fields == '' and len(addSystemContactFields) == 0 and len(addLinkedContactFields) == 0 and len(addSystemFields) == 0 and len(addLinkedAccountFields) == 0 and not addAll):
            raise Exception(
                'Please specify one or more entity or system fields, or specify \'addAll\'==True')

        if (len(addSystemFields) > 0):
            warnings.warn(
                "The addSystemFields parameter has been deprecated. Please use addSystemContactFields")
            for field in addSystemFields:
                if field not in addSystemContactFields:
                    addSystemContactFields.append(field)

        if (len(addLinkedContactFields) > 0 and entity != 'customObjects'):
            raise Exception(
                'Linked contact fields may only be included for CDO exports')

        if (len(addLinkedAccountFields) > 0 and entity != 'contacts'):
            raise Exception(
                'Linked account fields may only be included for contact exports')

        fieldStatement = {}

        if (entity == 'contacts' and leadScoreModelId > 0):
            fieldStatement[
                'Rating'] = '{{Contact.LeadScore.Model[%s].Rating}}' % leadScoreModelId
            fieldStatement[
                'Profile'] = '{{Contact.LeadScore.Model[%s].ProfileScore}}' % leadScoreModelId
            fieldStatement[
                'Engagement'] = '{{Contact.LeadScore.Model[%s].EngagementScore}}' % leadScoreModelId
        elif(entity != 'contacts' and leadScoreModelId > 0):
            raise TypeError(
                "Lead Scoring fields can only be included with contact exports")

        if (entity == 'activities'):
            if activityType in ('EmailSend', 'EmailOpen', 'EmailClickthrough', 'FormSubmit', 'Subscribe',
                                'Unsubscribe', 'Bounceback', 'WebVisit', 'PageView'):
                if len(addActivityFields) > 0:
                    for field in addActivityFields:
                        if field in system_fields.ACTIVITY_FIELDS[activityType]:
                            fieldStatement[field] = system_fields.ACTIVITY_FIELDS[
                                activityType][field]
                        else:
                            raise ValueError(
                                "Activity field not recognized: " + field)
                else:
                    fieldStatement = system_fields.ACTIVITY_FIELDS[
                        activityType]
            else:
                raise ValueError("Invalid activity type: " + activityType)
        else:

            if type(fields).__name__ == 'dict':
                fieldSet = self.GetFields(
                    entity=entity, fields=fields.values(), cdoID=cdoID)
            else:
                fieldSet = self.GetFields(
                    entity=entity, fields=fields, cdoID=cdoID)

            if len(addSystemFields) > 0:
                for field in addSystemContactFields:
                    if field in system_fields.CONTACT_SYSTEM_FIELDS:
                        fieldStatement[
                            field] = system_fields.CONTACT_SYSTEM_FIELDS[field]
                    else:
                        raise ValueError(
                            "System field not recognized: " + field)

            if len(fieldSet) > 0 or addAll:
                if type(fields).__name__ in ['list', 'str']:
                    for field in fieldSet:
                        if useInternalName:
                            fieldStatement[field['internalName']
                                           ] = field['statement']
                        else:
                            fieldStatement[field['name']] = field['statement']
                elif type(fields).__name__ == 'dict':
                    for field in fields:
                        for row in fieldSet:
                            if fields[field] == row['internalName'] or fields[field] == row['name']:
                                fieldStatement[field] = row['statement']
                                break
                elif fields == '' and addAll:
                    for field in fieldSet:
                        if useInternalName:
                            fieldStatement[field['internalName']
                                           ] = field['statement']
                        else:
                            fieldStatement[field['name']] = field['statement']

            else:
                raise Exception("No fields found")

        if len(addLinkedContactFields) > 0:

            linkedContactFields = self.CreateFieldStatement(
                entity='contacts', fields=addLinkedContactFields)

            for field in linkedContactFields:
                linkedContactFields[field] = linkedContactFields[field].replace(
                    '{{Contact.', '{{CustomObject[%s].Contact.') % cdoID
                fieldStatement.update(linkedContactFields)

        if len(addLinkedAccountFields) > 0:

            linkedAccountFields = self.CreateFieldStatement(
                entity='accounts', fields=addLinkedAccountFields)

            for field in linkedAccountFields:
                linkedAccountFields[field] = linkedAccountFields[
                    field].replace('{{Account.', '{{Contact.Account.')
                fieldStatement.update(linkedAccountFields)

        return fieldStatement

    def GetCdoId(self, cdoName):
        """
            Get a Custom Data Object's ID given it's name

            Arguments:

            * cdoName -- verbatim name of CDO; case insensitive

        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        cdoName = cdoName.replace(' ', '*')

        url = self.bulkBase + '/customobjects?q="name=' + str(cdoName) + '"'

        req = requests.get(url, auth=self.auth)

        if req.json()['totalResults'] == 1:
            cdoUri = req.json()['items'][0]['uri']
            cdoId = int(cdoUri.replace('/customObjects/', ''))
            return cdoId
        elif req.json()['totalResults'] > 1:
            raise Exception("Multiple CDOs with matching name")
            raise Exception("WTF?")
        else:
            raise Exception("No matching CDOs found")

    def getLeadScoreModelId(self, modelName):
        '''
            Returns model ID for a given lead score model
        '''
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        modelName = modelName.replace(' ', '*')

        url = self.bulkBase + '/contacts/scoring/models?q="name=' + modelName + "'"

        req = requests.get(url, auth=self.auth)

        if req.json()['totalResults'] == 1:
            modelId = req.json()['items'][0]['id']
            return modelId
        elif req.json()['totalResults'] > 1:
            raise Exception("Multiple models found")
        else:
            raise Exception("No matching models found")

    def FilterExists(self, name, existsType):
        """
            Given an object name, create an "EXISTS" statement (Eloqua Bulk equivalent of "in")

            Arguments:

            * name -- Name of object
            * existsType -- type of existence; one of ContactFilter, ContactList, ContactSegment, or AccountList

        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        name = name.replace(' ', '*')

        if existsType == 'ContactFilter':
            url = self.bulkBase + '/contacts/filters?q="name=' + name + '"'
        elif existsType == 'ContactSegment':
            url = self.bulkBase + '/contacts/segments?q="name=' + name + '"'
        elif existsType == 'ContactList':
            url = self.bulkBase + '/contacts/lists?q="name=' + name + '"'
        elif existsType == 'AccountList':
            url = self.bulkBase + '/accounts/lists?q="name=' + name + '"'
        else:
            raise Exception(
                "Please choose a valid 'existsType': 'ContactFilter', 'ContactSegment', 'ContactList', 'AccountList'")

        req = requests.get(url, auth=self.auth)

        if req.json()['totalResults'] == 1:
            filterStatement = "EXISTS('" + \
                req.json()['items'][0]['statement'] + "')"
            return filterStatement
        elif req.json()['totalResults'] > 1:
            raise Exception("Multiple " + existsType + "s found")
        else:
            raise Exception("No matching " + existsType + " found")

    def CreateSyncAction(self, action, destination='', status='', listName='', listType=''):
        '''
            Create a sync action to include in an import/export definition

            Arguments:

            * action --
            * destination --
            * status --
            * listName --
            * listType --
        '''
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if action == 'setStatus' and destination == '':
            raise Exception("must specify a destination for setStatus")

        if action in ['add', 'remove'] and destination == '' and listName == '':
            raise Exception(
                "for add and remove, must specify destination or listName")

        if listName != '' and listType not in ['accounts', 'contacts']:
            raise Exception(
                "if specifying listName, must also specify listType (accounts or contacts)")

        if listName != '' and listType != '' and destination == '':

            name = listName.replace(' ', '*')

            if listType == 'accounts':
                url = self.bulkBase + '/accounts/lists?q="name=' + name + '"'
            elif listType == 'contacts':
                url = self.bulkBase + '/contacts/lists?q="name=' + name + '"'

            req = requests.get(url, auth=self.auth)

            if req.json()['totalResults'] == 1:
                destination = req.json()['items'][0]['statement']
            elif req.json()['totalResults'] > 1:
                raise Exception("Multiple matching listName found")
            else:
                raise Exception("No matching listName found")

        syncAction = {}
        syncAction['action'] = action
        syncAction['destination'] = destination
        if action == 'setStatus':
            syncAction['status'] = status

        return syncAction

    def FilterDateRange(self, entity, field='', start='', end='', cdoID=0):
        '''
            Given an Eloqua date field, create a bounded or open date range filter

            Arguments:

            * entity -- one of: contacts, customObjects, accounts, activities
            * field -- field to filter by; must resolve to a date type field
            * start -- beginning of date range
            * end -- end of date range
            * cdoID -- identifier of specific CDO; required if entity = 'customObjects'; use method GetCdoId to retrieve

        '''
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if (start == '' and end == ''):
            raise ValueError(
                "Please enter at least one datetime value: start, end")

        if (field == '' and entity != 'activities'):
            raise ValueError(
                "Parameter 'field' is required for entity '" + entity + "'")

        if (entity == 'activities'):
            field = 'ActivityDate'

        if start != '':
            try:
                test1 = datetime.strptime(start, '%Y-%m-%d %H:%M:%S')
            except:
                raise ValueError(
                    "Invalid datetime format for 'start'; use 'YYYY-MM-DD hh:mm:ss'")
        if end != '':
            try:
                test2 = datetime.strptime(end, '%Y-%m-%d %H:%M:%S')
            except:
                raise ValueError(
                    "Invalid datetime format for 'end'; use 'YYYY-MM-DD hh:mm:ss'")

        if (entity != 'activities' and field not in ['createdAt', 'updatedAt']):
            fieldDef = self.GetFields(
                entity=entity, fields=[field], cdoID=cdoID)

            if (fieldDef[0]['dataType'] != 'date'):
                raise Exception("Field '" + field + "' is not a date field")
            fieldStatement = fieldDef[0]['statement']
        elif (entity == 'activities'):
            fieldStatement = system_fields.ACTIVITY_FIELDS[
                'CommonFields']['ActivityDate']
        else:
            fieldStatement = system_fields.CONTACT_SYSTEM_FIELDS[field]

        statement = ''

        if (start != ''):
            statement += " '" + fieldStatement + "' >= '" + start + "' "
        if (start != '' and end != ''):
            statement += ' AND '
        if (end != ''):
            statement += " '" + fieldStatement + "' <= '" + end + "' "

        return statement

    def CreateDef(self, defType, entity, fields, cdoID=0, filters='',
                  activityType='', defName=str(datetime.now()),
                  identifierFieldName='', isSyncTriggeredOnImport=False,
                  syncActions=[], updateRule='always'):
        """
            Create an import/export definition

            Arguments:

            * defType -- One of: 'imports', 'exports'
            * entity --  one of: contacts, customObjects, activities, or accounts
            * fields -- A dictionary of fields to export; see CreateFieldStatement
            * filters -- A valid Bulk filter statement
            * defName -- Export definition name; defaults to current datetime
            * identifierFieldName -- unique identified field; required for imports
            * isSyncTriggeredOnImport -- If True, begin sync upon post of import data
            * syncActions -- list of actions to take on contacts with sync import

        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if (defType not in ['imports', 'exports']):
            raise Exception(
                "Please choose a defType value: 'imports', 'exports'")

        if len(fields) == 0:
            raise Exception("Please specify at least one field to export")

        if (defType == 'imports' and len(fields) > 100):
            raise Exception(
                "Eloqua Bulk API only supports imports of up to 100 fields")

        if entity not in ['contacts', 'customObjects', 'accounts', 'activities']:
            raise Exception(
                "Please choose a valid 'entity' value: 'contacts', 'accounts', 'customObjects', 'activities'")

        if defType == 'exports' and len(syncActions) > 10:
            raise Exception(
                "may only specify up to 10 syncActions for an export")

        if defType == 'imports' and len(syncActions) > 1:
            raise Exception("may only specify 1 syncActions for an import")

        if entity not in ['contacts', 'customObjects', 'accounts'] and len(syncActions) > 0:
            raise Exception(
                "syncActions currently only supported in pyeloqua for contacts, customObjects, and accounts")

        if entity == 'customObjects':
            if cdoID == 0:
                raise Exception("Please specify a cdoID")
            if (len(fields) > 100):
                raise Exception(
                    "Eloqua Bulk API can only export 100 CDO fields at a time")
            url = self.bulkBase + '/customobjects/' + \
                str(cdoID) + '/' + defType

        if entity == 'contacts':
            if (len(fields) > 250):
                raise Exception(
                    "Eloqua Bulk API can only export 250 contact fields at a time")
            url = self.bulkBase + '/contacts/' + defType

        if entity == 'accounts':
            if len(fields) > 100:
                raise Exception(
                    "Eloqua Bulk API can only export 100 account fields at a time")
            url = self.bulkBase + '/accounts/' + defType

        if entity == 'activities':
            if len(fields) > 100:
                raise Exception(
                    "Eloqua Bulk API can only export 100 activity fields at a time")
            if activityType == '':
                raise Exception("Please specify an activity type")
            url = self.bulkBase + '/activities/' + defType

            if filters == '':
                filters = "'{{Activity.Type}}' = '" + activityType + "' "
            else:
                filters = "'{{Activity.Type}}' = '" + \
                    activityType + "' AND " + filters

        if (defType == 'exports'):

            if (len(filters) > 0):

                data = {'name': defName, 'filter': filters,
                        'fields': fields, 'syncActions': syncActions}

            else:

                data = {'name': defName, 'fields': fields,
                        'syncActions': syncActions}

        else:

            data = {'name': defName, 'fields': fields,
                    'identifierFieldName': identifierFieldName,
                    'isSyncTriggeredOnImport': isSyncTriggeredOnImport,
                    'syncActions': syncActions,
                    'updateRule': updateRule}

        req = requests.post(url, data=json.dumps(
            data), headers=POST_HEADERS, auth=self.auth)

        if req.status_code == 201:

            return req.json()

        else:

            raise Exception(req.json()['failures'][0])

    def GetDef(self, defURI='', defType=''):
        '''
            Get a single export/import definition or all

            Arguments:

            * defURI -- URI of pre-existing import/export definition; optional if selecting all
            * defType -- Definition type to retrieve; optional if defURI is provided
        '''
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        defs = []

        if defURI != '':
            url = self.bulkBase + defURI

            req = requests.get(url, auth=self.auth)

            if req.status_code == 200:
                defs.append(req.json())
            else:
                raise Exception(req.json()['failures'][0])
        elif defType != '':
            if defType in ['exports', 'imports']:
                hasMore = True
                offset = 0

                while hasMore:

                    url = self.bulkBase + defType + '?offset=' + str(offset)

                    req = requests.get(url, auth=self.auth)

                    defs.extend(req.json()['items'])

                    hasMore = req.json()['hasMore']

                    offset += 1000

            else:
                raise ValueError(
                    "defType must be one of the following: 'exports', 'imports'")
        else:
            raise ValueError(
                "Must specify one of the following: defURI, defType")

        return defs

    def CreateSync(self, defObject={}, defURI=''):
        """
            Create a sync for a pre-defined import/export

            Arguments:

            * defObject -- JSON object returned from CreateDef; optional if defURI is provided
            * defURI -- URI of pre-existing import/export definition; optional if defObject is provided

        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if ('uri' not in defObject):

            if (len(defURI) == 0):

                raise Exception("Must include a valid defObject or defURI")

            else:

                uri = defURI

        else:

            uri = defObject['uri']

        url = self.bulkBase + '/syncs'

        data = {'syncedInstanceUri': uri}

        req = requests.post(url, data=json.dumps(
            data), headers=POST_HEADERS, auth=self.auth)

        if req.status_code == 201:

            return req.json()

        else:

            raise Exception("Could not create sync: " + uri)

    def CheckSyncStatus(self, syncObject={}, syncURI='', timeout=500, interval=10):
        """
            Get the current status of an existing sync

            Arguments:

            * syncObject -- JSON object returned from CreateSync; optional if syncURI is provided
            * syncURI -- URI of pre-existing sync; optional if syncObject is provided
            * timeout -- seconds to check sync
            * interval -- wait time between checking status

        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')
        if ('uri' not in syncObject):
            if (len(syncURI) == 0):
                raise Exception("Must include a valid syncObject or syncURI")
            else:
                uri = syncURI
        else:
            uri = syncObject['uri']

        url = self.bulkBase + uri

        waitTime = 0
        notSynced = True
        while notSynced:
            req = requests.get(url, auth=self.auth)
            if req.status_code != 200:  # TODO: Fix this error handling
                warnings.warn(req.json())
            status = req.json()['status']
            if (status in ['success', 'warning', 'error']):
                return status
            elif (waitTime < timeout):
                waitTime += interval
                time.sleep(interval)
            else:
                raise Exception(
                    "Export not finished syncing after " + str(waitTime) + " seconds: " + uri)

    def GetSyncedRecordCount(self, defObject={}, defURI=''):
        '''
            Get number of synced records for an export

            Arguments:

            * defObject -- JSON object returned from CreateDef; optional if defURI is provided
            * defURI -- URI of pre-existing import/export definition; optional if defObject is provided

        '''
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if ('uri' not in defObject):
            if (len(defURI) == 0):
                raise Exception("Must include a valid defObject or defURI")
            else:
                uri = defURI
        else:
            uri = defObject['uri']

        url = self.bulkBase + uri + '/data?limit=1&offset=0'

        req = requests.get(url, auth=self.auth)

        totalResults = req.json()['totalResults']

        return totalResults

    def GetSyncRejectedRecords(self, syncObject={}, syncURI='', maxRecords=1000):
        '''
            Returns set of (at most) first 1000 records rejected from import

            Arguments:
            * syncObject -- JSON object returned from CreateSync; optional if syncURI is provided
            * syncURI -- URI of pre-existing sync; optional if syncObject is provided
            * maxRecords -- maximum number of records to query; capped at 1000; use 1 to just get # of rejects
        '''
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')

        if ('uri' not in syncObject):
            if (len(syncURI) == 0):
                raise Exception("Must include a valid syncObject or syncURI")
            else:
                uri = syncURI
        else:
            uri = syncObject['uri']

        if (maxRecords > 1000):
            raise ValueError("maxRecords must be <= 1000")

        url = self.bulkBase + uri + '/rejects?limit=' + str(maxRecords)

        req = requests.get(url, auth=self.auth)

        rejects = req.json()

        if (rejects['totalResults'] > 0):

            messages = []

            for row in rejects['items']:

                messages.append(row['message'])

            messageSummary = {}

            for row in messages:
                if row in messageSummary.keys():
                    messageSummary[row] += 1
                else:
                    messageSummary[row] = 1

            rejects['messages'] = messageSummary

        return rejects

    def GetSyncedData(self, defObject={}, defURI='', limit=50000, initOffset=0, retrieveLimit=1000000):
        """
            Retrieve data from a synced export

            Arguments:

            * defObject -- JSON object returned from CreateDef; optional if defURI is provided
            * defURI -- URI of pre-existing import/export definition; optional if defObject is provided
            * limit -- max number of records to retrieve (Eloqua max = 50,000) in one REST call; optional; different from retrieveLimit in that it will still pull all records, just in smaller batches
            * initOffset -- Starting offset to retrieve from; optional
            * retrieveLimit -- Max number of records to retrieve (total); optional
        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')
        if ('uri' not in defObject):
            if (len(defURI) == 0):
                raise Exception("Must include a valid defObject or defURI")
            else:
                uri = defURI
        else:
            uri = defObject['uri']

        offset = initOffset

        url = self.bulkBase + uri + '/data?'

        results = []

        hasMore = True

        while hasMore and len(results) < retrieveLimit:
            if (offset + limit) > retrieveLimit:
                limit = retrieveLimit - offset
            urlWhile = url + 'offset=' + str(offset) + '&limit=' + str(limit)
            req = requests.get(urlWhile, auth=self.auth)
            if 'items' in req.json():
                items = req.json()['items']
                for item in items:
                    results.append(item)
                offset += limit
            hasMore = req.json()['hasMore']

        return results

    def PostSyncData(self, data, defObject={}, defURI='', maxPost=20000, syncCount=80000, timeout=1000, interval=60):
        """
            Post data to an import definition and sync at regular intervals

            Arguments:

            * data -- list of data to be imported where each record is a dictionary
            * defObject -- JSON object returned from CreateDef; optional if defURI is provided
            * defURI -- URI of pre-existing import/export definition; optional if defObject is provided
            * maxPost -- max count of records to import at once
            * syncCount -- threshold for syncing posted data

        """
        warnings.warn('DEPRECATED: Please use the new Bulk class methods')
        if ('uri' not in defObject):
            if (len(defURI) == 0):
                raise Exception("Must include a valid defObject or defURI")
            else:
                uri = defURI
        else:
            uri = defObject['uri']

        if (maxPost > 20000):
            raise Exception(
                "It is not recommended to POST more than 20,000 records at a time. Please indicate a different maxPost value")

        if (syncCount > 80000):
            raise Exception(
                "It is recommended to sync at least every 80,000 records. Please indicate a different syncCount value")

        if (len(data) == 0):
            raise Exception("Input data length is 0")

        hasMore = True
        offset = 0
        syncOffset = 0
        sendSet = []
        dataLen = len(data)
        url = self.bulkBase + uri + '/data'
        syncSet = []
        sendSetLen = 0

        while (hasMore):
            for x in range(offset, min(offset + maxPost, dataLen), 1):
                sendSet.append(data[x])

            req = requests.post(url, data=json.dumps(sendSet, ensure_ascii=False).encode(
                'utf8'), headers=POST_HEADERS, auth=self.auth)

            sendSetLen += len(sendSet)

            if req.status_code in [204, 201]:

                syncOffset += maxPost

                if (syncOffset >= syncCount or offset + maxPost >= dataLen):
                    syncOffset = 0
                    importSync = self.CreateSync(
                        defObject=defObject, defURI=defURI)
                    syncStatus = self.CheckSyncStatus(
                        syncObject=importSync, timeout=timeout, interval=interval)
                    syncInfo = {"uri": importSync['uri'], 'status': syncStatus}
                    syncInfo['count'] = sendSetLen
                    syncInfo['rejectCount'] = self.GetSyncRejectedRecords(
                        syncObject=importSync, maxRecords=1)['totalResults']
                    syncSet.append(syncInfo)
                    sendSetLen = 0

                if offset + maxPost >= dataLen:
                    hasMore = False
                    # return syncSet
                else:
                    offset += maxPost
                    sendSet = []
            elif req.status_code == 400:
                raise Exception("400: Bad Request: ", req.json())
            elif req.status_code == 401:
                raise Exception("401: Unauthorized")
            elif req.status_code == 403:
                raise Exception("403: Forbidden")
            elif req.status_code == 404:
                raise Exception("404: Resource not found")
            elif req.status_code == 409:
                raise Exception("409: Conflict")
            elif req.status_code == 410:
                raise Exception("410: Resource no longer available")
            elif req.status_code == 413:
                raise Exception("413: Storage space exceeded")
            elif req.status_code == 500:
                raise Exception("500: Internal Eloqua server error")
            elif req.status_code == 503:
                raise Exception("503: Eloqua server timeout")
            else:
                # TODO: Fix this error handling
                raise Exception(req.status_code)

        return syncSet

    '''
        ###################################################
        Eloqua Forms
        ###################################################
    '''

    def GetForm(self, formId=0, formHtmlName='', formName=''):
        '''
            Retreive Eloqua form metadata

            Arguments:
            * formId -- ID key of Eloqua form
            * formHtmlName -- HTML name of Eloqua form
            * formName -- Display name of Eloqua form
        '''

        if (formId == 0 and formHtmlName == '' and formName == ''):
            raise ValueError(
                "Value required for one of: formId, formHtmlName, formName")
        if ((formId != 0 and (formHtmlName != '' or formName != '')) or (formHtmlName != '' and formName != '')):
            raise ValueError("More than one form identifier entered")

        if (formId != 0):
            url = self.restBase + '/assets/form/' + \
                str(formId) + '?depth=complete'
        else:
            url = self.restBase + '/assets/forms?depth=complete&search="' + \
                formName + formHtmlName + '"'

        req = requests.get(url, auth=self.auth)

        if (req.status_code == 200):
            form = req.json()
            # if ('elements' in form.keys()):
            #    formElem = form['elements']
            #    form = formElem[0]
            return form
        else:
            raise Exception("Form not found: " + str(formId))

    def ValidateFormFields(self, data, form):
        '''
            Given a single-record dictionary of data to submit, validate that all fields are present in the specified form

            Arguments:
            * data -- dictionary of data to post to an Eloqua form
            * form -- Output from GetForm function
        '''

        formFieldSet = form['elements']
        formFields = []
        formFieldsHtml = []
        formFieldsNotFound = []

        for row in formFieldSet:
            formFields.append(row['name'])
            formFieldsHtml.append(row['htmlName'])

        for row in data.keys():
            if row not in formFields:
                if row not in formFieldsHtml:
                    formFieldsNotFound.append(row)

        if (len(formFieldsNotFound) > 0):
            raise Exception(
                "Following fields not found on form: " + ", ".join(formFieldsNotFound))
        else:
            return 1

    def PostToForm(self, data, formId=0, formHtmlName='', formName=''):
        '''
            Post a dictionary of data to an Eloqua form

            Arguments:
            * data -- dictionary of data to submit; can be either a dict (single record), or list of dict (multiple records). Keys must match HTML form names
            * formId -- ID key of Eloqua form
            * formHtmlName -- HTML name of Eloqua form
            * formName -- Display name of Eloqua form
        '''

        form = self.GetForm(
            formId=formId, formHtmlName=formHtmlName, formName=formName)

        url = 'https://s' + str(self.siteId) + '.t.eloqua.com/e/f2'

        if (isinstance(data, dict)):
            data = [data]

        querystring = {'elqSiteID': self.siteId,
                       'elqFormName': form['htmlName']}

        successCount = 0
        failCount = 0

        for row in data:
            val = self.ValidateFormFields(data=row, form=form)
            req = requests.post(url, params=querystring, data=row)
            if (req.content == b'\r\n'):
                successCount += 1
            else:
                failCount += 1

        return {'success': successCount, 'failure': failCount}

    '''
        ###################################################
        REST Functions
        ###################################################
    '''

    def DeleteRecord(self, entity, id, cdoID=0):
        '''
            Deletes a given record. Cannot be undone.

            Arguments:
            * entity -- one of: contact, customObject, account
            * id -- ID of record to be deleted. It will be gone forever.
            * cdoID -- identifier of specific CDO; required if entity = 'customObject'; use method GetCdoId to retrieve
        '''

        if entity not in ['contact', 'customObject', 'account']:
            raise ValueError(
                "Please choose a valid 'entity' value: 'contact', 'account', 'customObject'")

        if entity == 'customObject' and cdoID == 0:
            raise ValueError("Please input a valid cdoID")

        if entity in ['contact', 'account']:
            uri = self.restBase + '/data/' + entity + '/' + str(id)
        else:
            uri = self.restBase + '/data/customObject/' + \
                str(cdoID) + '/instance/' + str(id)

        req = requests.delete(uri, auth=self.auth)

        return req.status_code

    def GetAssetSize(self, assetType, name):
        """
            Returns the count of contacts in a given shared list

            Arguments:
            * assetType -- one of: sharedList
            * name -- name of the asset
        """

        name1 = name.replace(' ', '*')

        if assetType not in ['sharedList']:
            raise ValueError(
                "Please choose a valid 'assetType' value: 'sharedList'")

        if assetType == 'sharedList':
            url = self.bulkBase + '/contacts/lists?q="name=' + name1 + '"'

        req = requests.get(url, auth=self.auth)

        if req.json()['totalResults'] == 1:
            count = req.json()['items'][0]['count']
        elif req.json()['totalResults'] > 1:
            for item in req.json()['items']:
                if item['name'] == name:
                    asset = item
                    break

            count = item['count']
        else:
            raise Exception("No matching " + assetType + " found")

        return int(count)

    def GetAsset(self, assetType, assetId):
        """
            Returns dict of asset info if found; None if not found

            Arguments:
            :param string assetType: list, filter, or segment
        """

        if assetType not in ['list', 'filter', 'segment']:
            raise ValueError(
                "Please choose a valid assetType: list, filter, segment")

        url = self.restBase + '/assets/contact/' + \
            assetType + '/' + str(assetId)

        req = requests.get(url, auth=self.auth)

        if req.status_code == 200:
            return req.json()
        else:
            return None



""" Eloqua.Bulk sync functions """

from copy import deepcopy
from json import dumps
from nose.tools import raises
from mock import patch, Mock

from pyeloqua import Bulk, EloquaBulkSyncTimeout

###############################################################################
# Constants
###############################################################################

POST_HEADERS = {'Content-Type': 'application/json'}

EXPORT_JOB_DEF = {
    "name": "test name",
    "fields": {
        "contactID": "{{Contact.Id}}",
        "createdAt": "{{Contact.CreatedAt}}",
        "updatedAt": "{{Contact.UpdatedAt}}",
        "isSubscribed": "{{Contact.Email.IsSubscribed}}",
        "isBounced": "{{Contact.Email.IsBounced}}",
        "emailFormat": "{{Contact.Email.Format}}"
    },
    "dataRetentionDuration": "PT12H",
    "uri": "/contacts/exports/1",
    "createdBy": "testuser",
    "createdAt": "2017-02-13T16:32:31.7020994Z",
    "updatedBy": "testuser",
    "updatedAt": "2017-02-13T16:32:31.7020994Z"
}

IMPORT_JOB_DEF = {
    "name": "test name",
    "fields": {
        "contactID": "{{Contact.Id}}",
        "createdAt": "{{Contact.CreatedAt}}",
        "updatedAt": "{{Contact.UpdatedAt}}",
        "isSubscribed": "{{Contact.Email.IsSubscribed}}",
        "isBounced": "{{Contact.Email.IsBounced}}",
        "emailFormat": "{{Contact.Email.Format}}"
    },
    "identifierFieldName": "contactID",
    "isSyncTriggeredOnImport": False,
    "dataRetentionDuration": "P7D",
    "isUpdatingMultipleMatchedRecords": False,
    "uri": "/contacts/imports/1",
    "createdBy": "testuser",
    "createdAt": "2017-02-13T16:38:13.3442894Z",
    "updatedBy": "testuser",
    "updatedAt": "2017-02-13T16:38:13.3442894Z"
}

SYNC_RESPONSE = {
    "syncedInstanceUri": "/contacts/exports/1",
    "status": "pending",
    "createdAt": "2015-09-25T18:08:32.3485942Z",
    "createdBy": "testuser",
    "uri": "/syncs/1"
}

SYNC_RESPONSE_ACTIVE = {
    "syncStartedAt": "2013-07-22T22:17:59.6730000Z",
    "status": "active",
    "createdAt": "2015-09-25T18:08:32.3485942Z",
    "createdBy": "testuser",
    "uri": "/syncs/1"
}

SYNC_RESPONSE_SUCCESS = {
    "syncStartedAt": "2013-07-22T22:17:59.6730000Z",
    "syncEndedAt": "2013-07-22T22:18:07.6430000Z",
    "status": "success",
    "createdAt": "2015-09-25T18:08:32.3485942Z",
    "createdBy": "testuser",
    "uri": "/syncs/1"
}


SYNC_RESPONSE_WARNING = {
    "syncStartedAt": "2013-07-22T22:17:59.6730000Z",
    "syncEndedAt": "2013-07-22T22:18:07.6430000Z",
    "status": "warning",
    "createdAt": "2015-09-25T18:08:32.3485942Z",
    "createdBy": "testuser",
    "uri": "/syncs/1"
}

SYNC_RESPONSE_ERROR = {
    "syncStartedAt": "2013-07-22T22:17:59.6730000Z",
    "syncEndedAt": "2013-07-22T22:18:07.6430000Z",
    "status": "error",
    "createdAt": "2015-09-25T18:08:32.3485942Z",
    "createdBy": "testuser",
    "uri": "/syncs/1"
}


###############################################################################
# You sync'd my data ship!
###############################################################################


@patch('pyeloqua.bulk.requests.post')
def test_start_sync_export(mock_post):
    """ start syncing an export - api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    bulk.start_sync()
    mock_post.assert_called_with(url=bulk.bulk_base + '/syncs',
                                 auth=bulk.auth,
                                 data=dumps(
                                     {
                                         "syncedInstanceUri": "/contacts/exports/1"
                                     },
                                 ensure_ascii=False).encode('utf8'),
                                 headers=POST_HEADERS)


@patch('pyeloqua.bulk.requests.post')
def test_start_sync_export_add(mock_post):
    """ start syncing an export - add to Bulk """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    bulk.start_sync()
    assert bulk.job_sync == SYNC_RESPONSE


@patch('pyeloqua.bulk.requests.post')
@raises(Exception)
def test_start_sync_export_nodef(mock_post):
    """ except when no job_def """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    bulk.start_sync()


###############################################################################
# check on the status of those syncs
###############################################################################

@patch('pyeloqua.bulk.requests.get')
def test_check_sync_call(mock_get):
    """ check on status of 'active' sync - api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_ACTIVE
    bulk.check_sync()
    mock_get.assert_called_with(url=bulk.bulk_base + '/syncs/1',
                                auth=bulk.auth)

@patch('pyeloqua.bulk.requests.get')
def test_check_sync_update(mock_get):
    """ check on status of 'active' sync - update job_sync """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_ACTIVE
    bulk.check_sync()
    assert bulk.job_sync == SYNC_RESPONSE_ACTIVE


@patch('pyeloqua.bulk.requests.get')
def test_check_sync_notfinished(mock_get):
    """ check on status of 'active' sync - return False """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_ACTIVE
    status = bulk.check_sync()
    assert status is False


@patch('pyeloqua.bulk.requests.get')
def test_check_sync_success(mock_get):
    """ check on status of 'success' sync - return True """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_SUCCESS
    status = bulk.check_sync()
    assert status is True


@patch('pyeloqua.bulk.requests.get')
def test_check_sync_warning(mock_get):
    """ check on status of 'warning' sync - return True """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_WARNING
    status = bulk.check_sync()
    assert status is True


@patch('pyeloqua.bulk.requests.get')
def test_check_sync_error(mock_get):
    """ check on status of 'error' sync - return True """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_ERROR
    status = bulk.check_sync()
    assert status is True


###############################################################################
# do all the sync needful
###############################################################################

@patch('pyeloqua.bulk.requests.post')
@patch('pyeloqua.bulk.requests.get')
def test_run_sync_success(mock_get, mock_post):
    """ fcn to run a sync end-to-end until finished - success """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_SUCCESS
    status = bulk.sync(sleeptime=0.01)
    assert status == 'success'


@patch('pyeloqua.bulk.requests.post')
@patch('pyeloqua.bulk.requests.get')
def test_run_sync_warning(mock_get, mock_post):
    """ fcn to run a sync end-to-end until finished - warning """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_WARNING
    status = bulk.sync(sleeptime=0.01)
    assert status == 'warning'


@patch('pyeloqua.bulk.requests.post')
@patch('pyeloqua.bulk.requests.get')
def test_run_sync_error(mock_get, mock_post):
    """ fcn to run a sync end-to-end until finished - error """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_ERROR
    status = bulk.sync(sleeptime=0.01)
    assert status == 'error'

@patch('pyeloqua.bulk.requests.post')
@patch('pyeloqua.bulk.requests.get')
@raises(EloquaBulkSyncTimeout)
def test_run_sync_timeout(mock_get, mock_post):
    """ fcn to run a sync end-to-end until finished - timeout """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=200)
    mock_post.return_value.json.return_value = deepcopy(SYNC_RESPONSE)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = SYNC_RESPONSE_ACTIVE
    bulk.sync(sleeptime=0.03, timeout=0.01)



""" Bulk API class """
from __future__ import print_function
from time import sleep
from datetime import datetime
from copy import deepcopy
from json import dumps, dump, load
import logging
import requests

from .pyeloqua import Eloqua
from .system_fields import ACTIVITY_FIELDS, CONTACT_SYSTEM_FIELDS, ACCOUNT_SYSTEM_FIELDS
from .error_handling import _elq_error_, EloquaBulkSyncTimeout, EloquaSyncError

############################################################################
# Constant definitions
############################################################################

POST_HEADERS = {'Content-Type': 'application/json'}

BLANK_JOB = {
    'filters': [],
    'fields': [],
    'job_type': None,
    'elq_object': None,
    'obj_id': None,
    'act_type': None,
    'options': {}
}

OBJECT_REQ_ID = ['customobjects', 'events']

OBJECT_REQ_TYPE = ['activities']

ELQ_OBJECTS = ['accounts', 'activities', 'contacts', 'customobjects',
               'emailaddresses', 'events']

logging.basicConfig(level=logging.WARN)

LOGGER = logging.getLogger(__name__)

############################################################################
# Bulk class
############################################################################


class Bulk(Eloqua):
    """ Extension for Bulk API operations """

    def __init__(self, username=None, password=None, company=None, test=False):
        """
        Initialize bulk class:

        Arguments:

        :param string username: Eloqua username
        :param string password: Eloqua password
        :param string company: Eloqua company instance
        :param bool test: Sets up test instance; does not connect to Eloqua
        :return: Bulk object
        """
        Eloqua.__init__(self, username, password, company, test)
        self.job = deepcopy(BLANK_JOB)
        self.job_def = {}
        self.job_sync = {}

    def reset(self):
        """ reset job """
        self.job = deepcopy(BLANK_JOB)
        self.job_def = {}
        self.job_sync = {}

    def _setup_(self, job_type, elq_object, obj_id=None, act_type=None):
        """
        setup a job

        Arguments:

        :param string job_type: 'imports' or 'exports'
        :param string elq_object: target Eloqua object
        :param int obj_id: parent ID for events or customobjects
        :param string act_type: Activity type
        """
        if elq_object not in ELQ_OBJECTS:
            raise Exception('invalid elq_object \'$s\'' % elq_object)
        # check if requires obj_id
        if elq_object in OBJECT_REQ_ID and obj_id is None:
            raise Exception('obj_id required for \'%s\'' % elq_object)
        if elq_object in OBJECT_REQ_TYPE and act_type is None:
            raise Exception('act_type required for \'%s\'' % elq_object)

        self.job['job_type'] = job_type
        self.job['elq_object'] = elq_object
        self.job['obj_id'] = obj_id
        self.job['act_type'] = act_type

        # If activity export, set export type
        if elq_object == 'activities':
            self.filter_equal('ActivityType', act_type)

    def imports(self, elq_object, obj_id=None, act_type=None):
        """
        setup a job with job_type == 'imports'

        Arguments:

        :param string elq_object: target Eloqua object
        :param int obj_id: parent ID for events or customobjects
        :param string act_type: Activity type
        """
        self._setup_(job_type='imports', elq_object=elq_object, obj_id=obj_id,
                     act_type=act_type)

    def exports(self, elq_object, obj_id=None, act_type=None):
        """
        setup a job with job_type == 'exports'

        Arguments:

        :param string elq_object: target Eloqua object
        :param int obj_id: parent ID for events or customobjects
        :param string act_type: Activity type
        """
        self._setup_(job_type='exports', elq_object=elq_object, obj_id=obj_id,
                     act_type=act_type)

    def write_job(self, path):
        """
        Write Bulk.job to a JSON file for later use

        :param str path: export path for job file
        """

        with open(path, 'w') as fopen:
            dump(self.job, fopen)

    def read_job(self, path):
        """
        Read a previously exported Bulk.job from a JSON file

        :param str path: import path for job file
        """

        with open(path, 'r') as fopen:
            self.job = load(fopen)


    ###########################################################################
    # Helper methods
    ###########################################################################

    def get_fields(self, elq_object=None, obj_id=None, act_type=None):
        """
        retrieve all fields for specified Eloqua object in job setup;
        useful if unsure what fields are available

        Arguments:

        :param string elq_object: target Eloqua object
        :param int obj_id: parent ID for events or customobjects
        :param string act_type: Activity type
        :return list: field definitions
        """
        # handle inputs vs Bulk.job
        if elq_object is None:
            elq_object = self.job['elq_object']
            obj_id = self.job['obj_id']
            act_type = self.job['act_type']
        # handle activity fields
        if elq_object == 'activities':
            return ACTIVITY_FIELDS[act_type]

        if elq_object in OBJECT_REQ_ID:
            url_base = self.bulk_base + '/{obj}/{id}/fields?limit=1000'.format(
                obj=elq_object,
                id=obj_id
            )
            url_base += '&offset={offset}'
        else:
            url_base = self.bulk_base + '/{obj}/fields?limit=1000'.format(
                obj=elq_object
            )
            url_base += '&offset={offset}'

        if elq_object == 'contacts':
            fields = deepcopy(CONTACT_SYSTEM_FIELDS)
        elif elq_object == 'accounts':
            fields = deepcopy(ACCOUNT_SYSTEM_FIELDS)
        else:
            fields = []

        has_more = True

        offset = 0

        while has_more:
            url = url_base.format(offset=offset)
            req = requests.get(url=url, auth=self.auth)
            _elq_error_(req)
            fields.extend(req.json()['items'])
            offset += 1
            has_more = req.json()['hasMore']

        return fields

    def add_fields(self, field_input=None):
        """
        retrieve all specified fields and add to job setup

        Arguments:

        :param list field_input: fields to add by DB name or Display Name
        """

        fields = self.get_fields()

        if field_input is None:
            self.job['fields'].extend(fields)
            return True

        fields_output = fields_intersect(fields, field_input)

        self.job['fields'].extend(fields_output)

    def add_linked_fields(self, lnk_obj, field_input):
        """
        add fields from linked objects

        :param string lnk_obj: linked object
        :param list field_input: fields to add by name
        """

        fields = self.get_fields(elq_object=lnk_obj)

        fields_output = fields_intersect(fields, field_input)

        for field in fields_output:
            if lnk_obj == 'contacts':
                if self.job['elq_object'] == 'customobjects':
                    field['statement'] = field['statement'].replace(
                        'Contact.', 'CustomObject[%s].Contact.' % self.job['obj_id'])
                elif self.job['elq_object'] == 'events':
                    field['statement'] = field['statement'].replace(
                        'Contact.', 'Event[%s].Contact.' % self.job['obj_id'])
            elif lnk_obj == 'accounts':
                field['statement'] = field['statement'].replace(
                    'Account.', 'Contact.Account.')

        self.job['fields'].extend(fields_output)

    def add_leadscore_fields(self, model_id=None, name=None):
        """
        add fields from a lead score model

        :param string name: name of lead score model
        :param int model_id: id of lead score model
        """

        if model_id is not None:
            url = self.bulk_base + \
                '/contacts/scoring/models/{0}'.format(model_id)
            req = requests.get(url=url, auth=self.auth)
            _elq_error_(req)

            self.job['fields'].extend(req.json()['fields'])
        elif name is not None:
            url = self.bulk_base + \
                '/contacts/scoring/models?q="name={name}"'.format(
                    name=name.replace(' ', '*'))
            req = requests.get(url=url, auth=self.auth)
            _elq_error_(req)

            self.job['fields'].extend(req.json()['items'][0]['fields'])
        else:
            raise Exception('model_id or name required')

    def asset_exists(self, asset, asset_id=None, name=None):
        """
        add filter statement for asset

        :param string asset: Eloqua asset type: lists, segments, filters
        :param int asset_id: id of asset
        :param string name: name of asset
        """

        exists_temp = " EXISTS('{statement}') "

        if asset_id is not None:
            url = self.bulk_base + '/{obj}/{asset}/{asset_id}'.format(
                obj=self.job['elq_object'],
                asset=asset,
                asset_id=asset_id
            )

            req = requests.get(url=url, auth=self.auth)

            _elq_error_(req)

            self.job['filters'].append(exists_temp.format(
                statement=req.json()['statement']))
        elif name is not None:
            url = self.bulk_base + '/{obj}/{asset}?q="name={name}"'.format(
                obj=self.job['elq_object'],
                asset=asset,
                name=name.replace(' ', '*')
            )

            req = requests.get(url=url, auth=self.auth)

            _elq_error_(req)

            self.job['filters'].append(exists_temp.format(
                statement=req.json()['items'][0]['statement']))
        else:
            raise Exception('asset_id or name required')

    def filter_date(self, field, start=None, end=None):
        """
        add a filter by start or end date

        :param string field: Field name on which to filter
        :param string start: Datetime string for date >=
        :param string end: Datetime string for date <=
        :param datetime start: Datetime object for date >=
        :param datetime end: Datetime object for date <=
        """

        field_stmt = fields_intersect(self.get_fields(), [field])[
            0]['statement']

        filter_str_start = None
        filter_str_end = None

        if start is not None:

            if isinstance(start, str):
                try:
                    datetime.strptime(start, '%Y-%m-%d %H:%M:%S')
                except ValueError:
                    raise Exception('invalid start value')
            elif isinstance(start, datetime):
                start = start.strftime('%Y-%m-%d %H:%M:%S')

            filter_str_start = " '{statement}' >= '{start}' ".format(
                statement=field_stmt,
                start=start
            )

        if end is not None:

            if isinstance(end, str):
                try:
                    datetime.strptime(end, '%Y-%m-%d %H:%M:%S')
                except ValueError:
                    raise Exception('invalid end value')
            elif isinstance(end, datetime):
                end = end.strftime('%Y-%m-%d %H:%M:%S')

            filter_str_end = " '{statement}' <= '{end}' ".format(
                statement=field_stmt,
                end=end
            )

        if filter_str_start is not None and filter_str_end is not None:
            filter_str = filter_str_start + 'AND' + filter_str_end
        elif filter_str_start is not None:
            filter_str = filter_str_start
        else:
            filter_str = filter_str_end

        self.job['filters'].append(filter_str)

    def filter_equal(self, field, value):
        """
        add filter statement for field equals value

        :param string field: Field name on which to filter
        :param string value: Field value
        """

        field_stmt = fields_intersect(self.get_fields(), [field])[
            0]['statement']

        filter_str = " '{statement}' = '{value}' ".format(
            statement=field_stmt,
            value=value
        )

        self.job['filters'].append(filter_str)

    def add_options(self, **kwargs):
        """
        add import/export options
        no client-side validation is performed; errors in server-side
        validation will result in a 400 status code
        """

        for opt in kwargs:
            self.job['options'][opt] = kwargs[opt]

    def add_syncaction(self, action, destination=None, status=None):
        """
        add sync actions
        No client-side validation

        :param str action: action to perform
        :param str destination: target,
        :param str status: set to status
        """

        sync_action = {
            'action': action,
            'destination': destination
        }

        if status is not None:
            sync_action['status'] = status

        if 'syncActions' not in self.job['options'].keys():
            self.job['options']['syncActions'] = []

        self.job['options']['syncActions'].append(sync_action)

    def add_syncaction_list(self, action, list_id=None, list_name=None):
        """
        add sync action based on shared list ID/Name

        :param int list_id: shared list ID
        :param str list_name: shared list name
        """

        if list_id is not None:
            destination = '{{ContactList[%s]}}' % list_id
        elif list_name is not None:
            url = self.bulk_base + '/{obj}/lists?q="name={list_name}"'.format(
                obj=self.job['elq_object'],
                list_name=list_name.replace(' ', '*')
            )

            req = requests.get(url=url, auth=self.auth)

            _elq_error_(req)

            destination = req.json()['items'][0]['statement']

        self.add_syncaction(action=action, destination=destination)

    def create_def(self, name):
        """
        create an import definition based on current object attributes

        :param str name: name of import/export definition
        """

        if self.job['elq_object'] in OBJECT_REQ_ID:
            url = self.bulk_base + '/{obj}/{id}/{job_type}'.format(
                obj=self.job['elq_object'],
                id=self.job['obj_id'],
                job_type=self.job['job_type']
            )
        else:
            url = self.bulk_base + '/{obj}/{job_type}'.format(
                obj=self.job['elq_object'],
                job_type=self.job['job_type']
            )

        req_data = {
            'name': name,
            'fields': {}
        }

        if self.job['filters']:
            req_data['filter'] = 'AND'.join(self.job['filters'])

        for field in self.job['fields']:
            if 'internalName' in field.keys():
                req_data['fields'][field['internalName']] = field['statement']
            else:
                req_data['fields'][field['name']] = field['statement']

        if self.job['options']:
            for option in self.job['options'].keys():
                req_data[option] = self.job['options'][option]

        req = requests.post(url=url, auth=self.auth, data=dumps(
            req_data, ensure_ascii=False).encode('utf8'), headers=POST_HEADERS)

        _elq_error_(req)

        self.job_def = req.json()

    def start_sync(self):
        """
        begin running a sync
        """

        req = requests.post(url=self.bulk_base + '/syncs',
                            auth=self.auth,
                            data=dumps({
                                'syncedInstanceUri': self.job_def['uri']
                            }, ensure_ascii=False).encode('utf8'),
                            headers=POST_HEADERS)

        _elq_error_(req)

        self.job_sync = req.json()

    def check_sync(self, uri=None):
        """
        check a pre-existing sync;
        update Bulk.job_sync
        :param str uri: pre-existing sync URI
        return True if finished, False if not finished
        """

        if uri is None:
            uri = self.job_sync['uri']

        req = requests.get(url=self.bulk_base + uri,
                           auth=self.auth)

        _elq_error_(req)

        self.job_sync = req.json()

        if self.job_sync['status'] in ['pending', 'active']:
            return False
        elif self.job_sync['status'] in ['success', 'warning', 'error']:
            return True

    def sync(self, timeout=600, sleeptime=5, start_new=True):
        """ run all sync actions; return final status of sync """

        if start_new is True:
            self.start_sync()

        finished = False

        time_elapsed = 0

        while finished is False:

            finished = self.check_sync()

            sleep(sleeptime)
            time_elapsed += sleeptime

            if time_elapsed >= timeout:
                raise EloquaBulkSyncTimeout(
                    'sync not finished after %s seconds' % time_elapsed)

        return self.job_sync['status']

    def handle_sync(self, **kwargs):
        """
        sync and handle errors; raise errors or log warnings

        :param **kwargs: keyword arguments for `Bulk.sync()`
        """
        status = self.sync(**kwargs)

        if status in ['warning', 'error']:

            logs = self.get_sync_logs()

            if status == 'warning':
                # keep processing if just a warning
                LOGGER.warning('status=warning')
                LOGGER.warning(logs)
            else:
                # throw exception otherwise
                LOGGER.error('status=error')
                raise EloquaSyncError(logs)

    def post_data(self, data):
        """ post data to bulk import """

        if self.job['job_type'] == 'exports':
            raise Exception('cannot POST data for an export')

        url = self.bulk_base + self.job_def['uri'] + '/data'

        req = requests.post(url=url, auth=self.auth, data=dumps(
            data, ensure_ascii=False).encode('utf8'), headers=POST_HEADERS)

        _elq_error_(req)

    def get_data(self, endpoint, max_recs=None, offset=0):
        """
        get data from a given endpoint
        this simplifies the looping process that would otherwise be repeated

        :param str endpoint: endpoint to be appended to bulk_base
        :param int max_recs: Max number of records to return
        """

        url_base = self.bulk_base + endpoint + '?limit={limit}&offset={offset}'

        limit, has_more = self._set_limit(max_recs, 0, True)

        return_data = []

        while has_more:

            url = url_base.format(limit=limit, offset=offset)

            req = requests.get(url=url, auth=self.auth)

            _elq_error_(req)

            if 'items' in req.json().keys():
                return_data.extend(req.json()['items'])

            offset += 1000

            if max_recs is None:

                has_more = req.json()['hasMore']

            limit, has_more = self._set_limit(max_recs, len(return_data),
                                              req.json()['hasMore'])

        return return_data


    def _set_limit(self, max_recs, row_ct, has_more):
        """  """

        # default limit return 1000
        # Used when pulling all data
        if max_recs is None:
            return 1000, has_more

        if (max_recs - row_ct) >= 1000:
            return 1000, True

        return (max_recs - row_ct), False


    def get_export_data(self, max_recs=None, offset=0):
        """
        retrieve all synced data for an export

        :param int limit: Max number of records to return from export
        """

        if self.job['job_type'] == 'imports':
            raise Exception('not an export')

        endpoint = self.job_def['uri'] + '/data'

        return_data = self.get_data(endpoint=endpoint, max_recs=max_recs, offset=offset)

        return return_data

    def get_sync_logs(self):
        """
        retrieve all sync logs
        """

        if self.job_sync == {}:
            raise Exception('no sync available')

        endpoint = self.job_sync['uri'] + '/logs'

        return_data = self.get_data(endpoint=endpoint)

        return return_data

    def get_sync_rejects(self):
        """
        retrieve all sync rejects
        """

        if self.job_sync == {}:
            raise Exception('no sync available')

        endpoint = self.job_sync['uri'] + '/rejects'

        return_data = self.get_data(endpoint=endpoint)

        return return_data


###############################################################################
# Helper functions
###############################################################################

def fields_intersect(field_set, field_input):
    """ Return field set given input set of names and input set of fields """
    fields_output = []

    for field_name in field_input:
        match = False
        for field in field_set:
            if 'internalName' in field.keys():
                if field_name == field['internalName'] or field_name == field['name']:
                    fields_output.append(field)
                    match = True
            else:
                if field_name == field['name']:
                    fields_output.append(field)
                    match = True
        if not match:
            raise Exception('field not found: %s' % field_name)

    return fields_output



""" Eloqua.Bulk job setup methods (imports/exports) """
from nose.tools import raises
from pyeloqua import Bulk

###############################################################################
# Methods to set up job
###############################################################################


def test_setup_type():
    """ setup set job_type """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'contacts')  # pylint: disable=W0212
    assert bulk.job['job_type'] == 'imports'


def test_setup_object():
    """ setup set object """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'contacts')  # pylint: disable=W0212
    assert bulk.job['elq_object'] == 'contacts'


@raises(Exception)
def test_setup_cdo_id_req():
    """ setup obj_id required for customobjects """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'customobjects')  # pylint: disable=W0212


def test_setup_cdo():
    """ setup sets obj_id """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'customobjects', 1)  # pylint: disable=W0212
    assert bulk.job['obj_id'] == 1


@raises(Exception)
def test_setup_event_id_req():
    """ setup obj_id required for events """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'events')  # pylint: disable=W0212


def test_setup_event():
    """ setup sets obj_id """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'events', 1)  # pylint: disable=W0212
    assert bulk.job['obj_id'] == 1


@raises(Exception)
def test_setup_actvty_type_req():
    """ setup act_type required for activities """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'activities')  # pylint: disable=W0212


def test_setup_actvty():
    """ setup sets obj_id """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'activities', # pylint: disable=W0212
                 act_type='EmailOpen')
    assert bulk.job['act_type'] == 'EmailOpen'


@raises(Exception)
def test_setup_bad_obj():
    """ setup obj_id required for events """
    bulk = Bulk(test=True)
    bulk._setup_('imports', 'bad')  # pylint: disable=W0212


def test_imports_job_type():
    """ imports() sets job_type """
    bulk = Bulk(test=True)
    bulk.imports('contacts')
    assert bulk.job['job_type'] == 'imports'


def test_exports_job_type():
    """ exports() sets job_type """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    assert bulk.job['job_type'] == 'exports'



""" Eloqua.Bulk class initialization """
from pyeloqua import Bulk, Eloqua

###############################################################################
# constants
###############################################################################

BLANK_JOB = {
    'filters': [],
    'fields': [],
    'job_type': None,
    'elq_object': None,
    'obj_id': None,
    'act_type': None,
    'options': {}
}

###############################################################################
# basic init
###############################################################################


def test_bulk_init():
    """ bulk class initializes """
    assert Bulk(test=True) is not None


def test_bulk_iselq():
    """ Is an Eloqua instance """
    bulk = Bulk(test=True)
    assert isinstance(bulk, Eloqua)


def test_bulk_hasjobdef():
    """ Has job_def which is dict """
    bulk = Bulk(test=True)
    assert isinstance(bulk.job_def, dict)

def test_bulk_hasjobsync():
    """ Has job_sync which is dict """
    bulk = Bulk(test=True)
    assert isinstance(bulk.job_sync, dict)

def test_bulk_hasjob():
    """ Has job_def which is instance of dict """
    bulk = Bulk(test=True)
    assert isinstance(bulk.job, dict)


def test_bulkdef_blankjob():
    """ BulkDef sets up 'job' """
    bulk = Bulk(test=True)
    assert bulk.job == BLANK_JOB


def test_reset_job():
    """ reset job params """
    bulk = Bulk(test=True)
    bulk.job['job_type'] = 'imports'
    bulk.reset()
    assert bulk.job == BLANK_JOB


def test_reset_job_def():
    """ reset job def """
    bulk = Bulk(test=True)
    bulk.job['job_type'] = 'imports'
    bulk.reset()
    assert bulk.job_def == {}

def test_reset_job_sync():
    """ reset job def """
    bulk = Bulk(test=True)
    bulk.job['job_type'] = 'imports'
    bulk.reset()
    assert bulk.job_sync == {}



""" POST data to a defined endpoint """

from json import dumps
from nose.tools import raises
from mock import patch, Mock

from pyeloqua import Bulk

###############################################################################
# Constants
###############################################################################

POST_HEADERS = {'Content-Type': 'application/json'}

IMPORT_JOB_DEF = {
    "name": "test name",
    "fields": {
        "contactID": "{{Contact.Id}}",
        "createdAt": "{{Contact.CreatedAt}}",
        "updatedAt": "{{Contact.UpdatedAt}}"
    },
    "identifierFieldName": "contactID",
    "isSyncTriggeredOnImport": False,
    "dataRetentionDuration": "P7D",
    "isUpdatingMultipleMatchedRecords": False,
    "uri": "/contacts/imports/1",
    "createdBy": "testuser",
    "createdAt": "2017-02-13T16:38:13.3442894Z",
    "updatedBy": "testuser",
    "updatedAt": "2017-02-13T16:38:13.3442894Z"
}

IMPORT_TEST_DATA = [
    {
        "contactID": "12345",
        "createdAt": "2017-01-01 00:00:00",
        "updatedAt": "2017-01-01 00:00:00"
    }
]

EXPORT_JOB_DEF = {
    "name": "test name",
    "fields": {
        "contactID": "{{Contact.Id}}",
        "createdAt": "{{Contact.CreatedAt}}",
        "updatedAt": "{{Contact.UpdatedAt}}",
        "isSubscribed": "{{Contact.Email.IsSubscribed}}",
        "isBounced": "{{Contact.Email.IsBounced}}",
        "emailFormat": "{{Contact.Email.Format}}"
    },
    "dataRetentionDuration": "PT12H",
    "uri": "/contacts/exports/1",
    "createdBy": "testuser",
    "createdAt": "2017-02-13T16:32:31.7020994Z",
    "updatedBy": "testuser",
    "updatedAt": "2017-02-13T16:32:31.7020994Z"
}

###############################################################################
# testing the things
###############################################################################


@patch('pyeloqua.bulk.requests.post')
def test_post_data_call(mock_post):
    """ post data to an import definition """
    bulk = Bulk(test=True)
    bulk.imports('contacts')
    bulk.job_def = IMPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=201)
    bulk.post_data(IMPORT_TEST_DATA)
    mock_post.assert_called_with(url=bulk.bulk_base + '/contacts/imports/1/data',
                                 auth=bulk.auth,
                                 data=dumps(IMPORT_TEST_DATA, ensure_ascii=False).encode('utf8'),
                                 headers=POST_HEADERS)


@patch('pyeloqua.bulk.requests.post')
@raises(Exception)
def test_post_data_except(mock_post):
    """ post data server error """
    bulk = Bulk(test=True)
    bulk.imports('contacts')
    bulk.job_def = IMPORT_JOB_DEF
    mock_post.return_value = Mock(ok=False, status_code=400)
    mock_post.return_value.json.return_value = {}
    bulk.post_data(IMPORT_TEST_DATA)


@patch('pyeloqua.bulk.requests.post')
@raises(Exception)
def test_post_data_except_import(mock_post):
    """ post data to an export definition """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_post.return_value = Mock(ok=True, status_code=201)
    bulk.post_data(IMPORT_TEST_DATA)



""" system field definitions (cannot be dynamically pulled but are available) """

CONTACT_SYSTEM_FIELDS = [
    {"name": "contactID", "statement": "{{Contact.Id}}"},
    {"name": "createdAt", "statement": "{{Contact.CreatedAt}}"},
    {"name": "updatedAt", "statement": "{{Contact.UpdatedAt}}"},
    {"name": "isSubscribed", "statement": "{{Contact.Email.IsSubscribed}}"},
    {"name": "isBounced", "statement": "{{Contact.Email.IsBounced}}"},
    {"name": "emailFormat", "statement": "{{Contact.Email.Format}}"}
]

ACCOUNT_SYSTEM_FIELDS = [
    {"name": "accountID", "statement": "{{Account.Id}}"},
    {"name": "createdAt", "statement": "{{Account.CreatedAt}}"},
    {"name": "updatedAt", "statement": "{{Account.UpdatedAt}}"}
]

ACTIVITY_FIELDS = {
    "EmailOpen": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "EmailAddress",
         "statement": "{{Activity.Field(EmailAddress)}}"},
        {"name": "ContactId", "statement": "{{Activity.Contact.Id}}"},
        {"name": "IpAddress", "statement": "{{Activity.Field(IpAddress)}}"},
        {"name": "VisitorId", "statement": "{{Activity.Visitor.Id}}"},
        {"name": "EmailRecipientId",
         "statement": "{{Activity.Field(EmailRecipientId)}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "SubjectLine",
         "statement": "{{Activity.Field(SubjectLine)}}"},
        {"name": "EmailWebLink",
         "statement": "{{Activity.Field(EmailWebLink)}}"},
        {"name": "VisitorExternalId", "statement": "{{Activity.Visitor.ExternalId}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"},
        {"name": "DeploymentId",
         "statement": "{{Activity.Field(EmailDeploymentId)}}"},
        {"name": "EmailSendType",
         "statement": "{{Activity.Field(EmailSendType)}}"}
    ],
    "EmailClickthrough": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "EmailAddress",
         "statement": "{{Activity.Field(EmailAddress)}}"},
        {"name": "ContactId", "statement": "{{Activity.Contact.Id}}"},
        {"name": "IpAddress", "statement": "{{Activity.Field(IpAddress)}}"},
        {"name": "VisitorId", "statement": "{{Activity.Visitor.Id}}"},
        {"name": "EmailRecipientId",
         "statement": "{{Activity.Field(EmailRecipientId)}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "SubjectLine",
         "statement": "{{Activity.Field(SubjectLine)}}"},
        {"name": "EmailWebLink",
         "statement": "{{Activity.Field(EmailWebLink)}}"},
        {"name": "EmailClickedThruLink",
         "statement": "{{Activity.Field(EmailClickedThruLink)}}"},
        {"name": "VisitorExternalId", "statement": "{{Activity.Visitor.ExternalId}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"},
        {"name": "DeploymentId",
         "statement": "{{Activity.Field(EmailDeploymentId)}}"},
        {"name": "EmailSendType",
         "statement": "{{Activity.Field(EmailSendType)}}"}
    ],
    "EmailSend": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "EmailAddress",
         "statement": "{{Activity.Field(EmailAddress)}}"},
        {"name": "ContactId", "statement": "{{Activity.Contact.Id}}"},
        {"name": "EmailRecipientId",
         "statement": "{{Activity.Field(EmailRecipientId)}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "SubjectLine",
         "statement": "{{Activity.Field(SubjectLine)}}"},
        {"name": "EmailWebLink",
         "statement": "{{Activity.Field(EmailWebLink)}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"},
        {"name": "DeploymentId",
         "statement": "{{Activity.Field(EmailDeploymentId)}}"},
        {"name": "EmailSendType",
         "statement": "{{Activity.Field(EmailSendType)}}"}
    ],
    "Subscribe": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "EmailAddress",
         "statement": "{{Activity.Field(EmailAddress)}}"},
        {"name": "EmailRecipientId",
         "statement": "{{Activity.Field(EmailRecipientId)}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"}
    ],
    "Unsubscribe": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "EmailAddress",
         "statement": "{{Activity.Field(EmailAddress)}}"},
        {"name": "EmailRecipientId",
         "statement": "{{Activity.Field(EmailRecipientId)}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"}
    ],
    "Bounceback": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "EmailAddress",
         "statement": "{{Activity.Field(EmailAddress)}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"},
        {"name": "EmailRecipientId", "statement": "{{Activity.Field(EmailRecipientId)}}"},
        {"name": "EmailDeploymentId", "statement": "{{Activity.Field(EmailDeploymentId)}}"},
        {"name": "SmtpErrorCode", "statement": "{{Activity.Field(SmtpErrorCode)}}"},
        {"name": "SmtpStatusCode", "statement": "{{Activity.Field(SmtpStatusCode)}}"},
        {"name": "SmtpMessage", "statement": "{{Activity.Field(SmtpMessage)}}"}
    ],
    "WebVisit": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "ContactId", "statement": "{{Activity.Contact.Id}}"},
        {"name": "VisitorId", "statement": "{{Activity.Visitor.Id}}"},
        {"name": "VisitorExternalId", "statement": "{{Activity.Visitor.ExternalId}}"},
        {"name": "ReferrerUrl",
         "statement": "{{Activity.Field(ReferrerUrl)}}"},
        {"name": "IpAddress", "statement": "{{Activity.Field(IpAddress)}}"},
        {"name": "NumberOfPages",
         "statement": "{{Activity.Field(NumberOfPages)}}"},
        {"name": "FirstPageViewUrl",
         "statement": "{{Activity.Field(FirstPageViewUrl)}}"},
        {"name": "Duration", "statement": "{{Activity.Field(Duration)}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"}
    ],
    "PageView": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "ContactId", "statement": "{{Activity.Contact.Id}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "VisitorId", "statement": "{{Activity.Visitor.Id}}"},
        {"name": "VisitorExternalId", "statement": "{{Activity.Visitor.ExternalId}}"},
        {"name": "WebVisitId", "statement": "{{Activity.Field(WebVisitId)}}"},
        {"name": "Url", "statement": "{{Activity.Field(Url)}}"},
        {"name": "ReferrerUrl",
         "statement": "{{Activity.Field(ReferrerUrl)}}"},
        {"name": "IpAddress", "statement": "{{Activity.Field(IpAddress)}}"},
        {"name": "IsWebTrackingOptedIn",
         "statement": "{{Activity.Field(IsWebTrackingOptedIn)}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"}
    ],
    "FormSubmit": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "ContactId", "statement": "{{Activity.Contact.Id}}"},
        {"name": "VisitorId", "statement": "{{Activity.Visitor.Id}}"},
        {"name": "VisitorExternalId", "statement": "{{Activity.Visitor.ExternalId}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetId", "statement": "{{Activity.Asset.Id}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "RawData", "statement": "{{Activity.Field(RawData)}}"},
        {"name": "CampaignId", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "ExternalId", "statement": "{{Activity.ExternalId}}"}
    ],
    "External": [
        {"name": "C_EmailAddress",
         "statement": "{{Activity.Contact.Field(C_EmailAddress)}}"},
        {"name": "CampaignID", "statement": "{{Activity.Campaign.Id}}"},
        {"name": "AssetName", "statement": "{{Activity.Asset.Name}}"},
        {"name": "AssetType", "statement": "{{Activity.Asset.Type}}"},
        {"name": "AssetDate", "statement": "{{Activity.CreatedAt}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"}
    ],
    "CommonFields": [
        {"name": "ActivityId", "statement": "{{Activity.Id}}"},
        {"name": "ActivityType", "statement": "{{Activity.Type}}"},
        {"name": "ActivityDate", "statement": "{{Activity.CreatedAt}}"}
    ]
}



""" generalized error handling """

def _elq_error_(request):
    """
    Deal with error codes on requests
    Eloqua serves 400 on validation errors; otherwise, general Exception
    is good enough

    :param Request request: object output from Requests
    """

    # deal with validation errors
    if request.status_code == 400:
        try:
            content = request.json()
        except Exception: # pylint: disable=broad-except
            content = request.text

        raise EloquaValidationError(request.status_code, content)

    if request.status_code >= 500:
        raise EloquaServerError(request.status_code)

    else:

        request.raise_for_status()

class EloquaSyncError(Exception):
    """ error on Eloqua Bulk Sync """
    pass

class EloquaBulkSyncTimeout(Exception):
    """ Exception class for Bulk sync timeouts """
    pass

class EloquaValidationError(Exception):
    """ Exception class for 400 errors """
    pass

class EloquaServerError(Exception):
    """ Exception class for 5xx errors """
    pass



""" Eloqua.Bulk job setup methods (filters) """

from datetime import datetime
from copy import deepcopy
from nose.tools import raises
from mock import patch, Mock

from pyeloqua import Bulk

###############################################################################
# Constants
###############################################################################

GOOD_LIST_ID = {
    "name": "Test List 1",
    "count": 1,
    "statement": "{{ContactList[1]}}",
    "uri": "/contacts/lists/1",
    "createdBy": "testuser",
    "createdAt": "2010-09-29T22:06:02.7430000Z",
    "updatedBy": "testuser",
    "updatedAt": "2013-09-24T13:13:41.4370000Z"
}

GOOD_LIST_NAME = {
    "items": [
        {
            "name": "Test List 1",
            "count": 1,
            "statement": "{{ContactList[1]}}",
            "uri": "/contacts/lists/1",
            "createdBy": "testuser",
            "createdAt": "2010-09-29T22:06:02.7430000Z",
            "updatedBy": "testuser",
            "updatedAt": "2013-09-24T13:13:41.4370000Z"
        }
    ],
    "totalResults": 1,
    "limit": 1000,
    "offset": 0,
    "count": 1,
    "hasMore": False
}

GOOD_CONTACT_FIELDS = {
    "items": [
        {
            "name": "Email Address",
            "internalName": "C_EmailAddress",
            "dataType": "emailAddress",
            "hasReadOnlyConstraint": False,
            "hasNotNullConstraint": False,
            "hasUniquenessConstraint": False,
            "statement": "{{Contact.Field(C_EmailAddress)}}",
            "uri": "/contacts/fields/1"
        },
        {
            "name": "First Name",
            "internalName": "C_FirstName",
            "dataType": "text",
            "hasReadOnlyConstraint": False,
            "hasNotNullConstraint": False,
            "hasUniquenessConstraint": False,
            "statement": "{{Contact.Field(C_FirstName)}}",
            "uri": "/contacts/fields/2"
        }
    ],
    "totalResults": 2,
    "limit": 1000,
    "offset": 0,
    "count": 2,
    "hasMore": False
}

###############################################################################
# Filter exists by asset
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_asset_exists_list_id(mock_get):
    """ add exists filter - Shared List by ID """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_LIST_ID)
    bulk.asset_exists(asset='lists', asset_id=1)
    assert bulk.job['filters'][0] == " EXISTS('{{ContactList[1]}}') "


@patch('pyeloqua.bulk.requests.get')
def test_asset_exists_list_id_call(mock_get):
    """ add exists filter - Shared List by ID, api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_LIST_ID)
    bulk.asset_exists(asset='lists', asset_id=1)
    mock_get.assert_called_with(url=bulk.bulk_base + '/contacts/lists/1',
                                auth=bulk.auth)


@patch('pyeloqua.bulk.requests.get')
def test_asset_exists_list_name(mock_get):
    """ add exists filter - Shared List by Name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_LIST_NAME)
    bulk.asset_exists(asset='lists', name='Test List 1')
    assert bulk.job['filters'][0] == " EXISTS('{{ContactList[1]}}') "


@patch('pyeloqua.bulk.requests.get')
def test_asset_exists_list_nm_call(mock_get):
    """ add exists filter - Shared List by Name, api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_LIST_NAME)
    bulk.asset_exists(asset='lists', name='Test List 1')
    mock_get.assert_called_with(url=bulk.bulk_base + '/contacts/lists?q="name=Test*List*1"',
                                auth=bulk.auth)


@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_asset_exists_list_none(mock_get):
    """ add exists filter - exception on no params """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_LIST_NAME)
    bulk.asset_exists(asset='lists')


###############################################################################
# Filter field by date range
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_filter_date_start(mock_get):
    """ add field filter by stating date """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    bulk.filter_date(field='createdAt', start='2017-01-01 00:00:00')
    assert bulk.job['filters'][
        0] == " '{{Contact.CreatedAt}}' >= '2017-01-01 00:00:00' "


@patch('pyeloqua.bulk.requests.get')
def test_filter_date_end(mock_get):
    """ add field filter by ending date """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    bulk.filter_date(field='createdAt', end='2017-01-01 00:00:00')
    assert bulk.job['filters'][
        0] == " '{{Contact.CreatedAt}}' <= '2017-01-01 00:00:00' "


@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_filter_date_nofield(mock_get):
    """ exception when bad field passed """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    bulk.filter_date(field='created', start='2017-01-01 00:00:00')


@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_filter_date_badstr_start(mock_get):
    """ exception when bad datetime string passed - start """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    bulk.filter_date(field='created', start='2017-01- 00:00:00')


@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_filter_date_badstr_end(mock_get):
    """ exception when bad datetime string passed - end """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    bulk.filter_date(field='created', end='2017-01- 00:00:00')


@patch('pyeloqua.bulk.requests.get')
def test_filter_datetime_start(mock_get):
    """ add field filter by stating datetime """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    dtime = datetime.strptime('2017-01-01 00:00:00', '%Y-%m-%d %H:%M:%S')
    bulk.filter_date(field='createdAt', start=dtime)
    assert bulk.job['filters'][
        0] == " '{{Contact.CreatedAt}}' >= '2017-01-01 00:00:00' "


@patch('pyeloqua.bulk.requests.get')
def test_filter_datetime_end(mock_get):
    """ add field filter by ending datetime """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    dtime = datetime.strptime('2017-01-01 00:00:00', '%Y-%m-%d %H:%M:%S')
    bulk.filter_date(field='createdAt', end=dtime)
    assert bulk.job['filters'][
        0] == " '{{Contact.CreatedAt}}' <= '2017-01-01 00:00:00' "


@patch('pyeloqua.bulk.requests.get')
def test_filter_datetime_both(mock_get):
    """ add field filter by stating datetime """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    dtime_start = datetime.strptime('2017-01-01 00:00:00', '%Y-%m-%d %H:%M:%S')
    dtime_end = datetime.strptime('2017-01-01 00:00:00', '%Y-%m-%d %H:%M:%S')
    bulk.filter_date(field='createdAt', start=dtime_start, end=dtime_end)
    assert bulk.job['filters'][
        0] == " '{{Contact.CreatedAt}}' >= '2017-01-01 00:00:00' AND '{{Contact.CreatedAt}}' <= '2017-01-01 00:00:00' "


###############################################################################
# Filter field equal to value
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_filter_value_equal(mock_get):
    """ add field filter by value equal """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_CONTACT_FIELDS)
    bulk.filter_equal(field='contactID', value='12345')
    assert bulk.job['filters'][0] == " '{{Contact.Id}}' = '12345' "



""" Eloqua class initialization """
from nose.tools import raises
from mock import patch, Mock

from requests.exceptions import HTTPError
from pyeloqua import Eloqua

###############################################################################
# constant definitions
###############################################################################

ELQ_LOGIN = {
    "site": {
        "id": 1234,
        "name": "test"
    },
    "user": {
        "id": 111,
        "username": "test",
        "displayName": "testing mctestface",
        "firstName": "testing",
        "lastName": "mctestface",
        "emailAddress": "test@test.com"
    },
    "urls": {
        "base": "https://secure.p01.eloqua.com",
        "apis": {
            "soap": {
                "standard":
                "https://secure.p01.eloqua.com/API/{version}/Service.svc",
                "dataTransfer":
                "https://secure.p01.eloqua.com/API/{version}/DataTransferService.svc",
                "email":
                "https://secure.p01.eloqua.com/API/{version}/EmailService.svc",
                "externalAction":
                "https://secure.p01.eloqua.com/API/{version}/ExternalActionService.svc"
            },
            "rest": {
                "standard": "https://secure.p01.eloqua.com/API/REST/{version}/",
                "bulk": "https://secure.p01.eloqua.com/API/Bulk/{version}/"
            }
        }
    }
}

###############################################################################
# exceptions
###############################################################################


@raises(Exception)
def test_elq_no_user():
    """ username required """
    Eloqua(company='test', password='test')


@raises(Exception)
def test_elq_no_company():
    """ company required """
    Eloqua(username='test', password='test')


@raises(Exception)
def test_elq_no_pwd():
    """ password required """
    Eloqua(company='test', username='test')


@patch('pyeloqua.pyeloqua.requests.get')
@raises(Exception)
def test_elq_invalid_auth(mock_get):
    """ exception raised on bad credentials """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = "Not authenticated."
    Eloqua(company='test', username='test', password='badtest')


@patch('pyeloqua.pyeloqua.requests.get')
@raises(HTTPError)
def test_elq_unknown_error(mock_get):
    """ exception raised on non-200 status code """
    mock_response = Mock()
    http_error = HTTPError()
    mock_response.raise_for_status.side_effect = http_error
    mock_get.return_value = mock_response
    Eloqua(company='test', username='test', password='test')

###############################################################################
# object values set
###############################################################################

@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_user(mock_get):
    """ set username """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.username == 'test'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_company(mock_get):
    """ set company """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.company == 'test'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_password(mock_get):
    """ set password """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.password == 'test'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_auth(mock_get):
    """ set auth tuple """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.auth == ('test\\test', 'test')


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_userid(mock_get):
    """ set user ID """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.userId == 111


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_user_id(mock_get):
    """ set user_id """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.user_id == 111


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_userdisplay(mock_get):
    """ set user display name """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.userDisplay == 'testing mctestface'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_user_display(mock_get):
    """ set user display name """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.user_display == 'testing mctestface'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_urlbase(mock_get):
    """ set URL base """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.urlBase == 'https://secure.p01.eloqua.com'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_url_base(mock_get):
    """ set url_base """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.url_base == 'https://secure.p01.eloqua.com'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_siteid(mock_get):
    """ set site ID """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.siteId == 1234


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_site_id(mock_get):
    """ set site_id """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.site_id == 1234


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_rest_base(mock_get):
    """ set REST API base """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.rest_base == 'https://secure.p01.eloqua.com/API/REST/2.0/'


@patch('pyeloqua.pyeloqua.requests.get')
def test_elq_set_bulk_base(mock_get):
    """ set Bulk API base """
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = ELQ_LOGIN
    elq = Eloqua(company='test', username='test', password='test')
    assert elq.bulk_base == 'https://secure.p01.eloqua.com/API/Bulk/2.0/'


def test_elq_test():
    """ test mode for authentication """
    elq = Eloqua(test=True)
    assert elq.auth == ('test\\test', 'test')

###############################################################################
# config options
###############################################################################

def test_elq_set_bulk_version():
    """ Change Bulk API version """
    elq = Eloqua(test=True)
    elq.set_bulk_version('1.2')
    assert elq.bulk_base == 'https://secure.p01.eloqua.com/API/BULK/1.2/'



""" pyeloqua """

from setuptools import setup, find_packages

__version__ = '0.5.7'

def readme():
    """ open readme for long_description """
    with open('README.md') as fle:
        return fle.read()

setup(
    name='pyeloqua',
    version=__version__,
    description='Python wrapper functions for Eloqua APIs',
    long_description=readme(),
    url='https://github.com/colemanja91/pyeloqua',
    author='Jeremiah Coleman',
    author_email='colemanja91@gmail.com',
    license='MIT',
    packages=find_packages(),
    install_requires=[
        'requests',
        'pbr',
        'six'
    ],
    zip_safe=False,
    keywords='eloqua marketing automation api bulk'
)



""" Eloqua.Bulk job setup methods (fields) """

from copy import deepcopy
from nose.tools import raises
from mock import patch, Mock

from pyeloqua import Bulk, ACTIVITY_FIELDS, CONTACT_SYSTEM_FIELDS, ACCOUNT_SYSTEM_FIELDS

###############################################################################
# Constants
###############################################################################

GOOD_FIELDS = {
    "items": [
        {
            "name": "Email Address",
            "internalName": "C_EmailAddress",
            "dataType": "emailAddress",
            "hasReadOnlyConstraint": False,
            "hasNotNullConstraint": False,
            "hasUniquenessConstraint": False,
            "statement": "{{Contact.Field(C_EmailAddress)}}",
            "uri": "/contacts/fields/1"
        },
        {
            "name": "First Name",
            "internalName": "C_FirstName",
            "dataType": "text",
            "hasReadOnlyConstraint": False,
            "hasNotNullConstraint": False,
            "hasUniquenessConstraint": False,
            "statement": "{{Contact.Field(C_FirstName)}}",
            "uri": "/contacts/fields/2"
        }
    ],
    "totalResults": 2,
    "limit": 1000,
    "offset": 0,
    "count": 2,
    "hasMore": False
}

ALL_CONTACT_FIELDS = []
ALL_CONTACT_FIELDS.extend(CONTACT_SYSTEM_FIELDS)
ALL_CONTACT_FIELDS.extend(GOOD_FIELDS['items'])

GOOD_FIELDS_ACCOUNT = {
    "items": [
        {
            "name": "Company Name",
            "internalName": "M_CompanyName",
            "dataType": "string",
            "hasReadOnlyConstraint": False,
            "hasNotNullConstraint": True,
            "hasUniquenessConstraint": False,
            "statement": "{{Account.Field(M_CompanyName)}}",
            "uri": "/accounts/fields/100085",
            "createdAt": "1900-01-01T05:00:00.0000000Z",
            "updatedBy": "Greg.Manetti",
            "updatedAt": "2012-10-30T17:59:21.7000000Z"
        },
        {
            "name": "Country",
            "internalName": "M_Country",
            "dataType": "string",
            "hasReadOnlyConstraint": False,
            "hasNotNullConstraint": False,
            "hasUniquenessConstraint": False,
            "statement": "{{Account.Field(M_Country)}}",
            "uri": "/accounts/fields/100088",
            "createdAt": "1900-01-01T05:00:00.0000000Z",
            "updatedAt": "1900-01-01T05:00:00.0000000Z"
        }
    ],
    "totalResults": 2,
    "limit": 1000,
    "offset": 0,
    "count": 2,
    "hasMore": False
}

ALL_ACCOUNT_FIELDS = []
ALL_ACCOUNT_FIELDS.extend(ACCOUNT_SYSTEM_FIELDS)
ALL_ACCOUNT_FIELDS.extend(GOOD_FIELDS_ACCOUNT['items'])

LINKED_ACCOUNT_FIELDS = [
    {
        "name": "Company Name",
        "internalName": "M_CompanyName",
        "dataType": "string",
        "hasReadOnlyConstraint": False,
        "hasNotNullConstraint": True,
        "hasUniquenessConstraint": False,
        "statement": "{{Contact.Account.Field(M_CompanyName)}}",
        "uri": "/accounts/fields/100085",
        "createdAt": "1900-01-01T05:00:00.0000000Z",
        "updatedBy": "Greg.Manetti",
        "updatedAt": "2012-10-30T17:59:21.7000000Z"
    },
    {
        "name": "Country",
        "internalName": "M_Country",
        "dataType": "string",
        "hasReadOnlyConstraint": False,
        "hasNotNullConstraint": False,
        "hasUniquenessConstraint": False,
        "statement": "{{Contact.Account.Field(M_Country)}}",
        "uri": "/accounts/fields/100088",
        "createdAt": "1900-01-01T05:00:00.0000000Z",
        "updatedAt": "1900-01-01T05:00:00.0000000Z"
    }
]

LINKED_CDO_CONTACT_FIELDS = [
    {
        "name": "Email Address",
        "internalName": "C_EmailAddress",
        "dataType": "emailAddress",
        "hasReadOnlyConstraint": False,
        "hasNotNullConstraint": False,
        "hasUniquenessConstraint": False,
        "statement": "{{CustomObject[1].Contact.Field(C_EmailAddress)}}",
        "uri": "/contacts/fields/1"
    },
    {
        "name": "First Name",
        "internalName": "C_FirstName",
        "dataType": "text",
        "hasReadOnlyConstraint": False,
        "hasNotNullConstraint": False,
        "hasUniquenessConstraint": False,
        "statement": "{{CustomObject[1].Contact.Field(C_FirstName)}}",
        "uri": "/contacts/fields/2"
    }
]

LINKED_EVENT_CONTACT_FIELDS = [
    {
        "name": "Email Address",
        "internalName": "C_EmailAddress",
        "dataType": "emailAddress",
        "hasReadOnlyConstraint": False,
        "hasNotNullConstraint": False,
        "hasUniquenessConstraint": False,
        "statement": "{{Event[1].Contact.Field(C_EmailAddress)}}",
        "uri": "/contacts/fields/1"
    },
    {
        "name": "First Name",
        "internalName": "C_FirstName",
        "dataType": "text",
        "hasReadOnlyConstraint": False,
        "hasNotNullConstraint": False,
        "hasUniquenessConstraint": False,
        "statement": "{{Event[1].Contact.Field(C_FirstName)}}",
        "uri": "/contacts/fields/2"
    }
]

LEADSCORE_MODEL_NAME = {
    "items": [
        {
            "name": "Model1",
            "status": "Active",
            "id": 1,
            "fields": [
                {
                    "name": "Rating",
                    "statement": "{{Contact.LeadScore.Model[1].Rating}}",
                    "dataType": "string"
                },
                {
                    "name": "ProfileScore",
                    "statement": "{{Contact.LeadScore.Model[1].ProfileScore}}",
                    "dataType": "number"
                },
                {
                    "name": "EngagementScore",
                    "statement": "{{Contact.LeadScore.Model[1].EngagementScore}}",
                    "dataType": "number"
                }
            ],
            "uri": "/contacts/scoring/models/1",
            "createdBy": "testuser",
            "updatedBy": "testuser",
            "createdAt": "2015-07-15T20:04:45.1600000Z",
            "updatedAt": "2015-09-16T05:09:09.3570000Z"
        }
    ],
    "totalResults": 1,
    "limit": 1000,
    "offset": 0,
    "count": 1,
    "hasMore": False
}

LEADSCORE_MODEL_NAME_NONE = {
    "items": [
        {}
    ],
    "totalResults": 0,
    "limit": 1000,
    "offset": 0,
    "count": 0,
    "hasMore": False
}

LEADSCORE_MODEL_ID = {
    "name": "Model1",
    "status": "Active",
    "id": 1,
    "fields": [
        {
            "name": "Rating",
            "statement": "{{Contact.LeadScore.Model[1].Rating}}",
            "dataType": "string"
        },
        {
            "name": "ProfileScore",
            "statement": "{{Contact.LeadScore.Model[1].ProfileScore}}",
            "dataType": "number"
        },
        {
            "name": "EngagementScore",
            "statement": "{{Contact.LeadScore.Model[1].EngagementScore}}",
            "dataType": "number"
        }
    ],
    "uri": "/contacts/scoring/models/1",
    "createdBy": "testuser",
    "updatedBy": "testuser",
    "createdAt": "2015-07-15T20:04:45.1600000Z",
    "updatedAt": "2015-09-16T05:09:09.3570000Z"
}

LEADSCORE_MODEL_FIELDS = [
    {
        "name": "Rating",
        "statement": "{{Contact.LeadScore.Model[1].Rating}}",
        "dataType": "string"
    },
    {
        "name": "ProfileScore",
        "statement": "{{Contact.LeadScore.Model[1].ProfileScore}}",
        "dataType": "number"
    },
    {
        "name": "EngagementScore",
        "statement": "{{Contact.LeadScore.Model[1].EngagementScore}}",
        "dataType": "number"
    }
]

###############################################################################
# Method to get all fields
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_get_fields_cntcts_call(mock_get):
    """ find all contact fields - correct call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.get_fields()
    url = bulk.bulk_base + '/contacts/fields?limit=1000&offset=0'
    mock_get.assert_any_call(url=url, auth=bulk.auth)


@patch('pyeloqua.bulk.requests.get')
def test_get_fields_events_call(mock_get):
    """ find all event fields """
    bulk = Bulk(test=True)
    bulk.exports('events', 1)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.get_fields()
    url = bulk.bulk_base + '/events/1/fields?limit=1000&offset=0'
    mock_get.assert_any_call(url=url, auth=bulk.auth)


@patch('pyeloqua.bulk.requests.get')
def test_get_fields_cntcts_return(mock_get):
    """ find all contact fields - return correct items """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    fields = bulk.get_fields()
    assert fields == ALL_CONTACT_FIELDS


def test_get_fields_actvty_return():
    """ find all contact fields - return correct items """
    bulk = Bulk(test=True)
    bulk.exports('activities', act_type='EmailOpen')
    fields = bulk.get_fields()
    assert fields == ACTIVITY_FIELDS['EmailOpen']


@patch('pyeloqua.bulk.requests.get')
def test_get_fields_specify(mock_get):
    """ specify elq_object to return fields for """
    bulk = Bulk(test=True)
    bulk.exports('accounts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    fields = bulk.get_fields(elq_object='contacts')
    assert fields == ALL_CONTACT_FIELDS

###############################################################################
# Method to get specified object fields
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_add_fields_db(mock_get):
    """ add contact fields to object by DB/API Name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    fields = ['C_EmailAddress', 'C_FirstName']
    bulk.add_fields(fields)
    assert bulk.job['fields'] == GOOD_FIELDS['items']


@patch('pyeloqua.bulk.requests.get')
def test_add_fields_display(mock_get):
    """ add contact fields to object by Display Name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    fields = ['Email Address', 'First Name']
    bulk.add_fields(fields)
    assert bulk.job['fields'] == GOOD_FIELDS['items']


@patch('pyeloqua.bulk.requests.get')
def test_add_fields_all(mock_get):
    """ add contact fields to object by Display Name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.add_fields()
    print(len(bulk.job['fields']))
    print(len(ALL_CONTACT_FIELDS))
    assert bulk.job['fields'] == ALL_CONTACT_FIELDS


@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_add_fields_notfound(mock_get):
    """ raise exception when field not found """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    fields = ['C_EmailAddress', 'C_FirstName', 'C_LastName']
    bulk.add_fields(fields)


def test_add_fields_actvty_all():
    """ add all activity fields """
    bulk = Bulk(test=True)
    bulk.exports('activities', act_type='EmailOpen')
    bulk.add_fields()
    assert bulk.job['fields'] == ACTIVITY_FIELDS['EmailOpen']


def test_add_fields_actvty():
    """ add some activity fields """
    bulk = Bulk(test=True)
    bulk.exports('activities', act_type='EmailOpen')
    fields = ['ActivityId', 'AssetId']
    bulk.add_fields(fields)
    assert len(bulk.job['fields']) == 2


@patch('pyeloqua.bulk.requests.get')
def test_cntct_system_fields_set(mock_get):
    """ add some contact system fields """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.add_fields(['contactID', 'createdAt'])
    assert len(bulk.job['fields']) == 2


@patch('pyeloqua.bulk.requests.get')
def test_accnt_system_fields_set(mock_get):
    """ add some account system fields """
    bulk = Bulk(test=True)
    bulk.exports('accounts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.add_fields(['accountID', 'createdAt'])
    assert len(bulk.job['fields']) == 2

###############################################################################
# Add linked record fields
# CDO -> Contact
# Contact -> Account
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_cntct_lnk_acct_fields(mock_get):
    """ add some linked account fields """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS_ACCOUNT)
    bulk.add_linked_fields('accounts', ['M_CompanyName', 'M_Country'])
    assert bulk.job['fields'] == LINKED_ACCOUNT_FIELDS


@patch('pyeloqua.bulk.requests.get')
def test_cdo_lnk_cntct_fields(mock_get):
    """ add some linked contact fields to cdo """
    bulk = Bulk(test=True)
    bulk.exports('customobjects', 1)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.add_linked_fields('contacts', ['C_EmailAddress', 'C_FirstName'])
    assert bulk.job['fields'] == LINKED_CDO_CONTACT_FIELDS


@patch('pyeloqua.bulk.requests.get')
def test_event_lnk_cntct_fields(mock_get):
    """ add some linked contact fields to event """
    bulk = Bulk(test=True)
    bulk.exports('events', 1)
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(GOOD_FIELDS)
    bulk.add_linked_fields('contacts', ['C_EmailAddress', 'C_FirstName'])
    assert bulk.job['fields'] == LINKED_EVENT_CONTACT_FIELDS

###############################################################################
# Lead Score Model fields
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_leadscore_fields_id(mock_get):
    """ add fields from a lead score model by model id """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(LEADSCORE_MODEL_ID)
    bulk.add_leadscore_fields(model_id=1)
    assert bulk.job['fields'] == LEADSCORE_MODEL_FIELDS

@patch('pyeloqua.bulk.requests.get')
def test_leadscore_fields_id_call(mock_get):
    """ lead score model by model id api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(LEADSCORE_MODEL_ID)
    bulk.add_leadscore_fields(model_id=1)
    mock_get.assert_called_with(url=bulk.bulk_base + '/contacts/scoring/models/1',
                                auth=bulk.auth)

@patch('pyeloqua.bulk.requests.get')
def test_leadscore_fields_name(mock_get):
    """ add fields from a lead score model by model name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(LEADSCORE_MODEL_NAME)
    bulk.add_leadscore_fields(name='Model 1')
    assert bulk.job['fields'] == LEADSCORE_MODEL_FIELDS

@patch('pyeloqua.bulk.requests.get')
def test_leadscore_fields_name_call(mock_get):
    """ lead score model by model name api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(LEADSCORE_MODEL_NAME)
    bulk.add_leadscore_fields(name='Model 1')
    mock_get.assert_called_with(url=bulk.bulk_base + '/contacts/scoring/models?q="name=Model*1"',
                                auth=bulk.auth)

@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_leadscore_fields_name_none(mock_get):
    """ add fields from a lead score model by model name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(LEADSCORE_MODEL_NAME_NONE)
    bulk.add_leadscore_fields(name='Model1')

@patch('pyeloqua.bulk.requests.get')
@raises(Exception)
def test_leadscore_fields_noparams(mock_get):
    """ add fields from a lead score model by model name """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(LEADSCORE_MODEL_NAME)
    bulk.add_leadscore_fields()



""" all methods of GETting data for Eloqua.Bulk """

from copy import deepcopy
from nose.tools import raises
from mock import patch, Mock

from pyeloqua import Bulk

###############################################################################
# Constants
###############################################################################

EXPORT_JOB_DEF = {
    "name": "test name",
    "fields": {
        "contactID": "{{Contact.Id}}",
        "createdAt": "{{Contact.CreatedAt}}",
        "updatedAt": "{{Contact.UpdatedAt}}",
        "isSubscribed": "{{Contact.Email.IsSubscribed}}",
        "isBounced": "{{Contact.Email.IsBounced}}",
        "emailFormat": "{{Contact.Email.Format}}"
    },
    "dataRetentionDuration": "PT12H",
    "uri": "/contacts/exports/1",
    "createdBy": "testuser",
    "createdAt": "2017-02-13T16:32:31.7020994Z",
    "updatedBy": "testuser",
    "updatedAt": "2017-02-13T16:32:31.7020994Z"
}

IMPORT_JOB_DEF = {
    "name": "test name",
    "fields": {
        "contactID": "{{Contact.Id}}",
        "createdAt": "{{Contact.CreatedAt}}",
        "updatedAt": "{{Contact.UpdatedAt}}",
        "isSubscribed": "{{Contact.Email.IsSubscribed}}",
        "isBounced": "{{Contact.Email.IsBounced}}",
        "emailFormat": "{{Contact.Email.Format}}"
    },
    "identifierFieldName": "contactID",
    "isSyncTriggeredOnImport": False,
    "dataRetentionDuration": "P7D",
    "isUpdatingMultipleMatchedRecords": False,
    "uri": "/contacts/imports/1",
    "createdBy": "testuser",
    "createdAt": "2017-02-13T16:38:13.3442894Z",
    "updatedBy": "testuser",
    "updatedAt": "2017-02-13T16:38:13.3442894Z"
}

RETURN_DATA = {
    "items": [
        {
            "contactID": "12345",
            "createdAt": "2017-01-01 00:00:00",
            "updatedAt": "2017-01-01 00:00:00"
        }
    ],
    "totalResults": 1,
    "limit": 1000,
    "offset": 0,
    "count": 1,
    "hasMore": False
}

RETURN_NORESULTS = {
    "totalResults": 1,
    "limit": 1000,
    "offset": 0,
    "count": 1,
    "hasMore": False
}

RETURN_SYNC_REJECTS = {
    "items": [
        {
            "fieldValues": {
                "contactID": "12345",
                "createdAt": "2017-01-01 00:00:00",
                "updatedAt": "2017-01-01 00:00:00"
            },
            "message": "Multiple matches.",
            "statusCode": "ELQ-00026",
            "recordIndex": 111,
            "invalidFields": [
                "emailAddress"
            ]
        },
    ],
    "totalResults": 1,
    "limit": 1000,
    "offset": 0,
    "count": 1,
    "hasMore": False
}

RETURN_SYNC_LOGS = {
    "items": [
        {
            "syncUri": "/syncs/845824",
            "count": 30000,
            "severity": "information",
            "statusCode": "ELQ-00130",
            "message": "Total records staged for import.",
            "createdAt": "2016-11-12T00:02:37.0770000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 0,
            "severity": "information",
            "statusCode": "ELQ-00137",
            "message": "Ready for data import processing.",
            "createdAt": "2016-11-12T00:02:37.0770000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 0,
            "severity": "information",
            "statusCode": "ELQ-00101",
            "message": "Sync processed for sync , resulting in Warning status.",
            "createdAt": "2016-11-12T00:02:52.0870000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 30000,
            "severity": "information",
            "statusCode": "ELQ-00001",
            "message": "Total records processed.",
            "createdAt": "2016-11-12T00:02:37.7000000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 21000,
            "severity": "warning",
            "statusCode": "ELQ-00026",
            "message": "Duplicate identifier.",
            "createdAt": "2016-11-12T00:02:46.2700000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 9000,
            "severity": "information",
            "statusCode": "ELQ-00003",
            "message": "Total records remaining after duplicates are rejected.",
            "createdAt": "2016-11-12T00:02:46.2700000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 0,
            "severity": "information",
            "statusCode": "ELQ-00004",
            "message": "Contacts created.",
            "createdAt": "2016-11-12T00:02:46.2700000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 9000,
            "severity": "information",
            "statusCode": "ELQ-00022",
            "message": "Contacts updated.",
            "createdAt": "2016-11-12T00:02:46.2700000Z"
        },
        {
            "syncUri": "/syncs/845824",
            "count": 0,
            "severity": "information",
            "statusCode": "ELQ-00071",
            "createdAt": "2016-11-12T00:02:46.2700000Z"
        }
    ],
    "totalResults": 9,
    "limit": 1000,
    "offset": 0,
    "count": 9,
    "hasMore": False
}

SYNC_RESPONSE_SUCCESS = {
    "syncStartedAt": "2013-07-22T22:17:59.6730000Z",
    "syncEndedAt": "2013-07-22T22:18:07.6430000Z",
    "status": "success",
    "createdAt": "2015-09-25T18:08:32.3485942Z",
    "createdBy": "testuser",
    "uri": "/syncs/1"
}

###############################################################################
# grab some datas
###############################################################################


@patch('pyeloqua.bulk.requests.get')
def test_get_data_call(mock_get):
    """ get data from an endpoint - api call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(RETURN_DATA)
    bulk.get_data(endpoint='/dummyurl')
    mock_get.assert_called_with(url=bulk.bulk_base + '/dummyurl?limit=1000&offset=0',
                                auth=bulk.auth)


@patch('pyeloqua.bulk.requests.get')
def test_get_data_return(mock_get):
    """ get data from an endpoint - return data """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(RETURN_DATA)
    return_data = bulk.get_data(endpoint='/dummyurl')
    assert return_data == RETURN_DATA['items']


@patch('pyeloqua.bulk.requests.get')
def test_get_data_nodata(mock_get):
    """ get data from an endpoint - no results data """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_get.return_value = Mock(ok=True, status_code=200)
    mock_get.return_value.json.return_value = deepcopy(RETURN_NORESULTS)
    return_data = bulk.get_data(endpoint='/dummyurl')
    assert return_data == []

###############################################################################
# get sunk'd export data
###############################################################################


@patch('pyeloqua.bulk.Bulk.get_data')
def test_get_export_data_call(mock_data):
    """ get data from a synced export - method call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_data.return_value = RETURN_DATA['items']
    bulk.get_export_data()
    mock_data.assert_called_with(endpoint='/contacts/exports/1/data',
                                 max_recs=None,
                                 offset=0)


@patch('pyeloqua.bulk.Bulk.get_data')
def test_get_export_data_rt(mock_data):
    """ get data from a synced export - return data """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_def = EXPORT_JOB_DEF
    mock_data.return_value = RETURN_DATA['items']
    data = bulk.get_export_data()
    assert data == RETURN_DATA['items']


@patch('pyeloqua.bulk.Bulk.get_data')
@raises(Exception)
def test_get_export_data_notexp(mock_data):
    """ get data from a synced export - exception """
    bulk = Bulk(test=True)
    bulk.imports('contacts')
    bulk.job_def = IMPORT_JOB_DEF
    mock_data.return_value = RETURN_DATA['items']
    bulk.get_export_data()


###############################################################################
# get sync logs
###############################################################################

@patch('pyeloqua.bulk.Bulk.get_data')
def test_get_sync_logs_call(mock_data):
    """ get logs from a sync - method call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE_SUCCESS
    mock_data.return_value = RETURN_SYNC_LOGS['items']
    bulk.get_sync_logs()
    mock_data.assert_called_with(endpoint='/syncs/1/logs')


@patch('pyeloqua.bulk.Bulk.get_data')
def test_get_sync_logs_rt(mock_data):
    """ get logs from a sync - return data """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE_SUCCESS
    mock_data.return_value = RETURN_SYNC_LOGS['items']
    data = bulk.get_sync_logs()
    assert data == RETURN_SYNC_LOGS['items']


@patch('pyeloqua.bulk.Bulk.get_data')
@raises(Exception)
def test_get_sync_logs_notexp(mock_data):
    """ get logs from a sync - exception """
    bulk = Bulk(test=True)
    bulk.imports('contacts')
    mock_data.return_value = RETURN_SYNC_LOGS['items']
    bulk.get_sync_logs()


###############################################################################
# get sync rejects
###############################################################################

@patch('pyeloqua.bulk.Bulk.get_data')
def test_get_sync_rejects_call(mock_data):
    """ get logs from a sync - method call """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE_SUCCESS
    mock_data.return_value = RETURN_SYNC_REJECTS['items']
    bulk.get_sync_rejects()
    mock_data.assert_called_with(endpoint='/syncs/1/rejects')


@patch('pyeloqua.bulk.Bulk.get_data')
def test_get_sync_rejects_rt(mock_data):
    """ get logs from a sync - return data """
    bulk = Bulk(test=True)
    bulk.exports('contacts')
    bulk.job_sync = SYNC_RESPONSE_SUCCESS
    mock_data.return_value = RETURN_SYNC_REJECTS['items']
    data = bulk.get_sync_rejects()
    assert data == RETURN_SYNC_REJECTS['items']


@patch('pyeloqua.bulk.Bulk.get_data')
@raises(Exception)
def test_get_sync_rejects_notexp(mock_data):
    """ get logs from a sync - exception """
    bulk = Bulk(test=True)
    bulk.imports('contacts')
    mock_data.return_value = RETURN_SYNC_REJECTS['items']
    bulk.get_sync_rejects()
